import matplotlib.pyplot as plt
import pytest

from pytest_mpl.plugin import switch_backend

from .helpers import bar, figure_test


class TestClass:

    @figure_test
    def test_hmatch_imatch_testclass(self):
        return bar([1, 2, 3, 4])

    @figure_test
    def test_hdiff_idiff_testclass(self):
        return bar([1, 3, 2, 4])


class TestClassWithSetupMethod:

    def setup_method(self, method):
        self.x = [4, 2, 3, 1]

    @figure_test
    def test_hmatch_imatch_testclasswithsetupmethod(self):
        return bar(self.x)

    @figure_test
    def test_hdiff_idiff_testclasswithsetupmethod(self):
        return bar(self.x)


class TestClassWithSetupClass:

    @classmethod
    def setup_class(cls):
        cls.x = [1, 0, 3, 2]

    @figure_test
    def test_hmatch_imatch_testclasswithsetupclass(self):
        return bar(self.x)

    @figure_test
    def test_hdiff_idiff_testclasswithsetupclass(self):
        return bar(self.x)


@pytest.fixture()
def figure_axes():
    with plt.style.context("classic", after_reset=True), switch_backend("agg"):
        fig = bar([3, 0, 4, 1])
    yield fig.gca()
    # Should not appear in test result
    fig.gca().plot([3, 0, 4, 1], c="yellow")


class TestClassWithFixture:
    @figure_test
    def test_hmatch_imatch_testclasswithfixture(self, figure_axes):
        figure_axes.plot([4, 1, 5, 2], c="red")
        return figure_axes.get_figure()


def generate_two_figures():
    for num, line in ((1, [1, 0, 1, 0]), (2, [0, 1, 0, 1])):
        plt.close(num)
        fig = plt.figure(num)
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(line)


class TestMultipleFigures:
    """
    Test figures are accessible, and can be passed to individual functions.

    See https://github.com/matplotlib/pytest-mpl/issues/133
    """
    @classmethod
    def setup_class(cls):
        # setting style and backend inside decorator has no effect
        # because figures are generated outside the test function
        with plt.style.context("classic", after_reset=True), switch_backend("agg"):
            generate_two_figures()
            cls.figs = [plt.figure(i) for i in (1, 2)]

    @figure_test
    def test_hmatch_imatch_multiplefigures_first(self):
        return self.figs[0]

    @figure_test
    def test_hmatch_imatch_multiplefigures_second(self):
        return self.figs[1]



import os
import shutil
import pathlib
import tempfile
import subprocess

from docutils import nodes
from sphinx.util.docutils import SphinxRole
from sphinx.util.osutil import canon_path

REPO_ROOT = pathlib.Path(__file__).parent.parent
TEST_FILE = REPO_ROOT / "tests" / "subtests" / "test_subtest.py"
SAMPLE_DIR = "sample"


def run_pytest(test_name):

    # Create generated samples directory
    sample_dir_abs = pathlib.Path(__file__).parent / SAMPLE_DIR
    if not sample_dir_abs.exists():
        os.mkdir(sample_dir_abs)

    # Form path to current sample
    dest = sample_dir_abs / test_name
    if dest.exists():
        return dest  # skip if already generated

    # Generate the current sample
    tmp_dir = tempfile.mkdtemp()
    command = f"python -m pytest {TEST_FILE}::{test_name} -v --mpl --basetemp={tmp_dir}"
    subprocess.run(command, shell=True, check=True)

    # Find the name of the directory the sample is within
    # (directory name is sometimes truncated)
    src = next(filter(
        lambda x: x.name[:-1] in test_name,
        pathlib.Path(tmp_dir).glob("*0")
    )) / "results"
    shutil.copytree(src, dest)

    return dest


class SummaryButtons(nodes.General, nodes.Inline, nodes.TextElement):
    pass


class SummaryRole(SphinxRole):
    def run(self):
        node = SummaryButtons(name=self.text)
        return [node], []


def move_summaries(app, *args, **kwargs):
    gen_sample_dir = pathlib.Path(__file__).parent / SAMPLE_DIR
    out_sample_dir = pathlib.Path(app.outdir) / SAMPLE_DIR
    if out_sample_dir.exists():
        shutil.rmtree(out_sample_dir)
    shutil.copytree(gen_sample_dir, out_sample_dir)


def html_visit_summary(self, node):

    test_name = str(node["name"])
    out = run_pytest(test_name)

    classes = (
        "sd-sphinx-override sd-btn sd-text-wrap sd-btn-{importance} "
        "sd-shadow-sm sd-me-2 reference internal"
    )
    button = (
        '<a class="{classes}" href="{href}" style="margin-right: 0.5rem;">{label}</a>'
    )

    summary_types = {
        "HTML": "fig_comparison.html",
        "Basic HTML": "fig_comparison_basic.html",
        "JSON": "results.json",
    }

    current_filename = self.builder.current_docname + self.builder.out_suffix
    current_dir = pathlib.PurePath(current_filename).parent
    first_button = True
    for label, file in summary_types.items():
        if (out / file).exists():
            importance = "primary" if first_button else "secondary"
            self.body.append(button.format(
                classes=classes.format(importance=importance),
                href=canon_path((current_dir / SAMPLE_DIR / test_name / file).as_posix()),
                label=label,
            ))
            first_button = False

    raise nodes.SkipNode


def skip(self, node):
    raise nodes.SkipNode


def setup(app):
    app.connect("build-finished", move_summaries)
    app.add_node(
        SummaryButtons,
        html=(html_visit_summary, None),
        latex=(skip, None),
        text=(skip, None),
        man=(skip, None),
        texinfo=(skip, None),
    )
    app.add_role("summary", SummaryRole())
    return {"parallel_read_safe": True, "parallel_write_safe": True}



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import datetime
from packaging.version import Version

from pytest_mpl import __version__

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'pytest-mpl'
author = 'Thomas Robitaille'
copyright = '{}, {}'.format(datetime.datetime.now().year, author)

release = __version__
pytest_mpl_version = Version(__version__)
is_release = not (pytest_mpl_version.is_prerelease or pytest_mpl_version.is_devrelease)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sample_summaries',
    'sphinx.ext.intersphinx',
    'sphinx_design',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

intersphinx_mapping = {
    "matplotlib": ("https://matplotlib.org/stable", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "mpl_sphinx_theme"
html_theme_options = {
    "navbar_links": "absolute",
    "show_prev_next": False,
    "logo": {"link": "https://matplotlib.org/stable/",
             "image_light": "images/logo_light.svg",
             "image_dark": "images/logo_dark.svg"},
    "collapse_navigation": False,
}
html_sidebars = {
    "**": ["mpl_third_party_sidebar.html", "sidebar-nav-bs.html"]
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']



import pytest

from .helpers import plot

# ### Specialized tests


# Tolerance: high to force image match
@pytest.mark.image
@pytest.mark.hash
@pytest.mark.mpl_image_compare(tolerance=200, savefig_kwargs={'metadata': {'Software': None}})
def test_hdiff_imatch_tolerance():
    return plot([1, 2, 3, 4], linestyle='--')


# Tolerance: non-default to verify option recorded in JSON
@pytest.mark.image
@pytest.mark.hash
@pytest.mark.mpl_image_compare(tolerance=3, savefig_kwargs={'metadata': {'Software': None}})
def test_hdiff_idiff_tolerance():
    return plot([1, 2, 3, 4], linestyle='--')


# Savefig kwargs
@pytest.mark.image
@pytest.mark.hash
@pytest.mark.mpl_image_compare(savefig_kwargs={'facecolor': 'r', 'metadata': {'Software': None}})
def test_hdiff_imatch_savefig():
    return plot([1, 2, 3, 4])


# TODO: Implement these path altering tests later
# # Different baseline directory
# # TODO: Test with a remote `baseline_dir`
# @pytest.mark.mpl_image_compare(baseline_dir='baseline/other')
# def test_hdiff_imatch_baselinedir():
#     return plot([4, 2, 1, 4])
#
#
# # Different filename
# @pytest.mark.mpl_image_compare(filename='test_hdiff_imatch_filename_other.png')
# def test_hdiff_imatch_filename():
#     return plot([4, 2, 1, 4])
#
#
# # Different hash library
# @pytest.mark.mpl_image_compare(hash_library='hashes/other/other.json')
# def test_hdiff_imatch_hashlibrary():
#     return plot([4, 2, 1, 4])


# Different style
@pytest.mark.image
@pytest.mark.hash
@pytest.mark.mpl_image_compare(style='fivethirtyeight', savefig_kwargs={'metadata': {'Software': None}})
def test_hdiff_imatch_style():
    return plot([4, 2, 1, 4])


# Remove text
@pytest.mark.image
@pytest.mark.hash
@pytest.mark.mpl_image_compare(remove_text=True, savefig_kwargs={'metadata': {'Software': None}})
def test_hdiff_imatch_removetext():
    return plot([4, 2, 1, 4])



import pytest
from packaging.version import Version

pytest_plugins = ["pytester"]

if Version(pytest.__version__) < Version("6.2.0"):
    @pytest.fixture
    def pytester(testdir):
        return testdir



import os
import sys
import shutil

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    cached_property = property

from jinja2 import Environment, PackageLoader, select_autoescape

__all__ = ['generate_summary_html', 'generate_summary_basic_html']


class Results:
    """
    Data for the whole test run, used for providing data to the template.

    Parameters
    ----------
    results : dict
        The `pytest_mpl.plugin.ImageComparison._test_results` object.
    title : str
        Value for HTML <title>.
    """
    def __init__(self, results, title="Image comparison"):
        self.title = title  # HTML <title>

        # Generate sorted list of results
        self.cards = []
        pad = len(str(len(results.items())))  # maximum length of a result index
        for collect_n, (name, item) in enumerate(results.items()):
            card_id = str(collect_n).zfill(pad)  # zero pad for alphanumerical sorting
            self.cards += [Result(name, item, card_id)]
        self.cards = sorted(self.cards, key=lambda i: i.indexes['status'], reverse=True)

    @cached_property
    def statistics(self):
        """Generate a dictionary of summary statistics."""
        stats = {'passed': 0, 'failed': 0, 'passed_baseline': 0,
                 'failed_baseline': 0, 'skipped': 0}
        for test in self.cards:
            if test.status == 'passed':
                stats['passed'] += 1
                if test.image_status != 'match':
                    stats['failed_baseline'] += 1
            elif test.status == 'failed':
                stats['failed'] += 1
                if test.image_status == 'match':
                    stats['passed_baseline'] += 1
            elif test.status == 'skipped':
                stats['skipped'] += 1
        return stats

    @cached_property
    def image_comparison(self):
        """Whether at least one image comparison test or generation was performed."""
        for result in self.cards:
            if result.image_status:
                return True
        return False

    @cached_property
    def hash_comparison(self):
        """Whether at least one hash comparison test or generation was performed."""
        for result in self.cards:
            if result.hash_status:
                return True
        return False


class Result:
    """
    Result data for a single image test, used for providing data to the template.

    Parameters
    ----------
    name : str
        Full name of the test including modules.
    item : dict
        Dictionary of summary results for a test in
        `pytest_mpl.plugin.ImageComparison._test_results`.
    id : str
        The test number in order collected. Numbers must be
        zero padded due to alphanumerical sorting.
    """
    def __init__(self, name, item, id):
        # Make the summary dictionary available as attributes
        self.__dict__ = item

        # Sort index for collection order
        self.id = id

        # Name of test with module and test function together and separate
        self.full_name = name
        without_parametrized_name = name.split("[")[0]
        self.module = ".".join(without_parametrized_name.split(".")[:-1])
        self.name = name[len(self.module)+1:]

        # Additional classes to add to the result card
        self.classes = [f'{k}-{str(v).lower()}' for k, v in [
            ('overall', self.status),
            ('image', self.image_status),
            ('hash', self.hash_status),
        ]]

    @cached_property
    def indexes(self):
        """Dictionary with strings optimized for sorting."""
        return {'status': self._status_sort, 'rms': self._rms_sort}

    @property
    def _status_sort(self):
        """Status number. Higher means more issues."""
        s = 0
        if self.status == 'failed':
            s += 10
        if self.image_status == 'diff':
            s += 3
        elif self.image_status == 'missing':
            s += 4
        if self.hash_status == 'diff':
            s += 1
        elif self.hash_status == 'missing':
            s += 5
        return f"{s:02.0f}"

    @property
    def _rms_sort(self):
        """RMS to 3 d.p. for sorting."""
        if self.image_status == 'match':
            return "000000"
        elif self.image_status == 'diff':
            if self.rms is None:  # Shape mismatch
                return "999999"
            # RMS will be in [0, 255]
            return f"{(self.rms + 2) * 1000:06.0f}"
        else:  # Missing baseline image
            return "000001"

    @cached_property
    def rms_str(self):
        """RMS to show in template."""
        if self.image_status == 'match':
            return '< tolerance'  # self.rms is None
        elif self.image_status == 'diff':
            return self.rms
        else:  # Missing baseline image
            return 'None'

    @property
    def badges(self):
        """Additional badges to show beside overall status badge."""
        for test_type, status_getter in [('image', image_status_msg), ('hash', hash_status_msg)]:
            status = getattr(self, f'{test_type}_status')
            if (
                    (status == 'missing') or
                    (self.status == 'failed' and status == 'match') or
                    (self.status == 'passed' and status == 'diff')
            ):  # Only show if different to overall status
                yield {'status': status, 'svg': test_type, 'tooltip': status_getter(status)}


def status_class(status):
    """Status to Bootstrap class."""
    status = status.split('-')[-1]  # e.g. "overall-passed" -> "passed"
    classes = {
        'passed': 'success',
        'failed': 'danger',
        'skipped': 'warning',
        'match': 'success',
        'diff': 'danger',
        'missing': 'warning',
        'generated': 'warning',
    }
    return classes[status]


def image_status_msg(status):
    """Image status to status message."""
    messages = {
        'match': 'Baseline image matches',
        'diff': 'Baseline image differs',
        'missing': 'Baseline image not found',
        'generated': 'Baseline image was generated',
    }
    return messages[status]


def hash_status_msg(status):
    """Hash status to status message."""
    messages = {
        'match': 'Baseline hash matches',
        'diff': 'Baseline hash differs',
        'missing': 'Baseline hash not found',
        'generated': 'Baseline hash was generated',
    }
    return messages[status]


def generate_summary_html(results, results_dir, hash_library=None):
    """Generate the HTML summary.

    Parameters
    ----------
    results : dict
        The `pytest_mpl.plugin.ImageComparison._test_results` object.
    results_dir : Path
        Path to the output directory.
    hash_library : str, optional, default=None
        Filename of the generated hash library at the root of `results_dir`.
        Will be linked to in HTML if not None.
    """

    # Initialize Jinja
    env = Environment(
        loader=PackageLoader("pytest_mpl.summary.html"),
        autoescape=select_autoescape()
    )

    # Register additional Jinja filters
    env.filters["status_class"] = status_class
    env.filters["image_status_msg"] = image_status_msg
    env.filters["hash_status_msg"] = hash_status_msg

    # Render HTML starting from the base template
    template = env.get_template("base.html")
    html = template.render(results=Results(results), hash_library=hash_library)

    # Write files
    for file in ['styles.css', 'extra.js', 'hash.svg', 'image.svg']:
        path = os.path.join(os.path.dirname(__file__), 'templates', file)
        shutil.copy(path, results_dir / file)
    html_file = results_dir / 'fig_comparison.html'
    with open(html_file, 'w') as f:
        f.write(html + '\n')

    return html_file


def generate_summary_basic_html(results, results_dir, hash_library=None):
    """Generate the basic HTML summary.

    Parameters
    ----------
    results : dict
        The `pytest_mpl.plugin.ImageComparison._test_results` object.
    results_dir : Path
        Path to the output directory.
    hash_library : str, optional, default=None
        Filename of the generated hash library at the root of `results_dir`.
        Will be linked to in HTML if not None.
    """

    # Initialize Jinja
    env = Environment(
        loader=PackageLoader("pytest_mpl.summary.html"),
        autoescape=select_autoescape()
    )

    # Render HTML starting from the base template
    template = env.get_template("basic.html")
    html = template.render(results=Results(results), hash_library=hash_library)

    # Write files
    html_file = results_dir / 'fig_comparison_basic.html'
    with open(html_file, 'w') as f:
        f.write(html + '\n')

    return html_file



import matplotlib
import matplotlib.pyplot as plt
import pytest
from helpers import pytester_path, skip_if_format_unsupported
from packaging.version import Version
from PIL import Image

MPL_VERSION = Version(matplotlib.__version__)

METADATA = {
    "png": {"Software": None},
    "pdf": {"Creator": None, "Producer": None, "CreationDate": None},
    "eps": {"Creator": "test"},
    "svg": {"Date": None},
}


def test_multiple_cli_flags(pytester):
    result = pytester.runpytest("--mpl", "--mpl-deterministic", "--mpl-no-deterministic")
    result.stderr.fnmatch_lines(
        ["*ValueError: Only one of `--mpl-deterministic` and `--mpl-no-deterministic` can be set.*"]
    )


def test_warning(pytester):
    path = pytester_path(pytester)
    hash_library = path / "hash_library.json"
    kwarg = f"hash_library=r'{hash_library}'"
    pytester.makepyfile(
        f"""
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare({kwarg})
        def test_mpl():
            fig, ax = plt.subplots()
            ax.plot([1, 3, 2])
            return fig
        """
    )
    result = pytester.runpytest(f"--mpl-generate-hash-library={hash_library}")
    result.stdout.fnmatch_lines(["*FutureWarning: deterministic option not set*"])
    result.assert_outcomes(failed=1)


@pytest.mark.parametrize("file_format", ["eps", "pdf", "png", "svg"])
@pytest.mark.parametrize(
    "ini, cli, kwarg, success_expected",
    [
        ("true", "", None, True),
        ("false", "--mpl-deterministic", None, True),
        ("true", "--mpl-no-deterministic", None, False),
        ("", "--mpl-no-deterministic", True, True),
        ("true", "", False, False),
    ],
)
@pytest.mark.skipif(MPL_VERSION < Version("3.3.0"), reason="Test unsupported: Default metadata is different in MPL<3.3")
def test_config(pytester, file_format, ini, cli, kwarg, success_expected):
    skip_if_format_unsupported(file_format, using_hashes=True)

    path = pytester_path(pytester)
    baseline_dir = path / "baseline"
    hash_library = path / "hash_library.json"

    ini = f"mpl-deterministic = {ini}" if ini else ""
    pytester.makeini(
        f"""
        [pytest]
        mpl-hash-library = {hash_library}
        {ini}
        """
    )

    kwarg = f", deterministic={kwarg}" if isinstance(kwarg, bool) else ""
    pytester.makepyfile(
        f"""
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare(savefig_kwargs={{'format': '{file_format}'}}{kwarg})
        def test_mpl():
            fig, ax = plt.subplots()
            ax.plot([1, 2, 3])
            return fig
        """
    )

    # Generate baseline hashes
    assert not hash_library.exists()
    pytester.runpytest(
        f"--mpl-generate-path={baseline_dir}",
        f"--mpl-generate-hash-library={hash_library}",
        cli,
    )
    assert hash_library.exists()
    baseline_image = baseline_dir / f"test_mpl.{file_format}"
    assert baseline_image.exists()
    deterministic_metadata = METADATA[file_format]

    if file_format == "svg":  # The only format that is reliably non-deterministic between runs
        result = pytester.runpytest("--mpl", f"--mpl-baseline-path={baseline_dir}", cli)
        if success_expected:
            result.assert_outcomes(passed=1)
        else:
            result.assert_outcomes(failed=1)

    elif file_format == "pdf":
        with open(baseline_image, "rb") as fp:
            file = str(fp.read())
        for metadata_key in deterministic_metadata.keys():
            key_in_file = fr"/{metadata_key}" in file
            if success_expected:  # metadata keys should not be in the file
                assert not key_in_file
            else:
                assert key_in_file

    else:  # "eps" or "png"
        with Image.open(str(baseline_image)) as image:
            actual_metadata = image.info
        for k, expected in deterministic_metadata.items():
            actual = actual_metadata.get(k, None)
            if success_expected:  # metadata keys should not be in the file
                if expected is None:
                    assert actual is None
                else:
                    assert actual == expected
            else:  # metadata keys should still be in the file
                if expected is None:
                    assert actual is not None
                else:
                    assert actual != expected



import unittest

from .helpers import bar, figure_test


class TestCase(unittest.TestCase):

    @figure_test
    def test_hmatch_imatch_testclass(self):
        return bar([1, 2, 3, 4])

    @figure_test
    def test_hdiff_idiff_testclass(self):
        return bar([1, 3, 2, 4])


class TestCaseWithSetUp(unittest.TestCase):

    def setUp(self):
        self.x = [4, 2, 3, 1]

    @figure_test
    def test_hmatch_imatch_testcasewithsetup(self):
        return bar(self.x)

    @figure_test
    def test_hdiff_idiff_testcasewithsetup(self):
        return bar(self.x)


class TestCaseWithSetUpClass(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.x = [4, 3, 2, 4]

    @figure_test
    def test_hmatch_imatch_testcasewithsetupclass(self):
        return bar(self.x)

    @figure_test
    def test_hdiff_idiff_testcasewithsetupclass(self):
        return bar(self.x)



import json

import pytest
from helpers import pytester_path


@pytest.mark.parametrize(
    "ini, cli, expected",
    [
        ("json", None, {"json"}),
        ("json", "html", {"html"}),
        ("basic-html", "json", {"json"}),
        (None, "json,basic-html,html", {"json", "basic-html", "html"}),
    ],
)
def test_config(pytester, ini, cli, expected):
    path = pytester_path(pytester)
    ini = f"mpl-generate-summary = {ini}" if ini else ""
    pytester.makeini(
        f"""
        [pytest]
        mpl-results-path = {path}
        {ini}
        """
    )
    pytester.makepyfile(
        """
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare
        def test_mpl():
            fig, ax = plt.subplots()
            ax.plot([1, 2, 3])
            return fig
        """
    )
    cli = f"--mpl-generate-summary={cli}" if cli else ""
    result = pytester.runpytest("--mpl", cli)
    result.assert_outcomes(failed=1)

    json_summary = path / "results.json"
    if "json" in expected:
        with open(json_summary) as fp:
            results = json.load(fp)
            assert "test_config.test_mpl" in results
    else:
        assert not json_summary.exists()

    html_summary = path / "fig_comparison.html"
    if "html" in expected:
        with open(html_summary) as fp:
            raw = fp.read()
        assert "bootstrap" in raw
        assert "test_config.test_mpl" in raw
    else:
        assert not html_summary.exists()

    basic_html_summary = path / "fig_comparison_basic.html"
    if "basic-html" in expected:
        with open(basic_html_summary) as fp:
            raw = fp.read()
        assert "bootstrap" not in raw
        assert "test_config.test_mpl" in raw
    else:
        assert not basic_html_summary.exists()



import pytest


@pytest.mark.parametrize(
    "ini, cli, kwarg, expected",
    [
        ("backend1", None, None, "backend1"),
        ("backend1", "backend2", None, "backend2"),
        ("backend1", "backend2", "backend3", "backend3"),
    ],
)
def test_config(pytester, ini, cli, kwarg, expected):
    ini = f"mpl-default-backend = {ini}" if ini else ""
    pytester.makeini(
        f"""
        [pytest]
        {ini}
        """
    )
    kwarg = f"backend='{kwarg}'" if kwarg else ""
    pytester.makepyfile(
        f"""
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare({kwarg})
        def test_mpl():
            fig, ax = plt.subplots()
            ax.plot([1, 2, 3])
            return fig
        """
    )
    cli = f"--mpl-default-backend={cli}" if cli else ""
    result = pytester.runpytest("--mpl", cli)
    result.assert_outcomes(failed=1)
    result.stdout.re_match_lines([f".*(ModuleNotFound|Value)Error: .*{expected}.*"])



from pathlib import Path

import pytest
from helpers import pytester_path


@pytest.mark.parametrize(
    "ini, cli, enabled_expected",
    [
        (None, None, False),
        (True, None, True),
        (False, None, False),
        (False, True, True),
        (True, True, True),
    ],
)
def test_config(pytester, ini, cli, enabled_expected):
    path = pytester_path(pytester)
    ini = f"mpl-results-always = {ini}" if ini else ""
    pytester.makeini(
        f"""
        [pytest]
        mpl-default-style = fivethirtyeight
        mpl-baseline-path = {Path(__file__).parent / "baseline" / "2.0.x"}
        mpl-results-path = {path}
        {ini}
        """
    )
    pytester.makepyfile(
        """
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare
        def test_base_style():
            fig, ax = plt.subplots()
            ax.plot([1, 2, 3])
            return fig
        """
    )
    cli = "--mpl-results-always" if cli else ""
    result = pytester.runpytest("--mpl", cli)
    result.assert_outcomes(passed=1)
    assert (path / "test_config.test_base_style" / "result.png").exists() == enabled_expected



from functools import wraps

import matplotlib.pyplot as plt
import numpy as np
import pytest


def plot(line, **kwargs):
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot(line, **kwargs)
    return fig


def bar(height, **kwargs):
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    x = np.arange(len(height))
    ax.bar(x, height, align="center", **kwargs)
    return fig


def figure_test(test_function, **kwargs):
    @pytest.mark.image
    @pytest.mark.hash
    @pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}}, **kwargs)
    @wraps(test_function)
    def test_wrapper(*args, **kwargs):
        return test_function(*args, **kwargs)
    return test_wrapper



import pytest
from helpers import pytester_path


@pytest.mark.parametrize(
    "ini, cli, kwarg, expected",
    [
        ("sty1", None, None, "sty1"),
        ("sty1", "sty2", None, "sty2"),
        ("sty1", "sty2", "sty3", "sty3"),
    ],
)
def test_config(pytester, ini, cli, kwarg, expected):
    ini = "mpl-default-style = " + ini if ini else ""
    pytester.makeini(
        f"""
        [pytest]
        mpl-baseline-path = {pytester_path(pytester)}
        {ini}
        """
    )
    kwarg = f"style='{kwarg}'" if kwarg else ""
    pytester.makepyfile(
        f"""
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare({kwarg})
        def test_mpl():
            fig, ax = plt.subplots()
            ax.plot([1, 2, 3])
            return fig
        """
    )

    cli = "--mpl-default-style=" + cli if cli else ""
    result = pytester.runpytest("--mpl", cli)
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines([f"*OSError: *'{expected}'*"])



import pytest
from helpers import pytester_path


@pytest.mark.parametrize(
    "ini, cli, expected",
    [
        ("dir1", None, "dir1"),
        ("dir1", "dir2", "dir2"),
        (None, "dir2", "dir2"),
    ],
)
def test_config(pytester, ini, cli, expected):
    ini = f"mpl-results-path = {ini}" if ini else ""
    pytester.makeini(
        f"""
        [pytest]
        {ini}
        """
    )
    pytester.makepyfile(
        """
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare
        def test_mpl():
            fig, ax = plt.subplots()
            ax.plot([1, 2, 3])
            return fig
        """
    )
    cli = f"--mpl-results-path={cli}" if cli else ""
    result = pytester.runpytest("--mpl", cli)
    result.assert_outcomes(failed=1)
    assert (pytester_path(pytester) / expected / "test_config.test_mpl" / "result.png").exists()



import os
import re
import json
from pathlib import Path

from PIL import Image, ImageDraw

__all__ = ['diff_summary', 'assert_existence', 'patch_summary', 'apply_regex',
           'remove_specific_hashes', 'transform_hashes', 'transform_images']


class MatchError(Exception):
    pass


def diff_summary(baseline, result, baseline_hash_library=None, result_hash_library=None,
                 generating_hashes=False):
    """Diff a pytest-mpl summary dictionary.

    Parameters
    ----------
    baseline : dict
        Baseline pytest-mpl summary.
    result : dict
        Generated result pytest-mpl summary.
    baseline_hash_library : Path, optional, default=None
        Path to the baseline hash library.
        Baseline hashes in the baseline summary are updated to these values
        to handle different Matplotlib versions.
    result_hash_library : Path, optional, default=None
        Path to the "baseline" image hash library.
        Result hashes in the baseline summary are updated to these values
        to handle different Matplotlib versions.
    generating_hashes : bool, optional, default=False
        Whether `--mpl-generate-hash-library` was specified and
        both of `--mpl-hash-library` and `hash_library=` were not.
    """
    if baseline_hash_library and baseline_hash_library.exists():
        # Load "correct" baseline hashes
        with open(baseline_hash_library, 'r') as f:
            baseline_hash_library = json.load(f)
    else:
        baseline_hash_library = {}
    if result_hash_library and result_hash_library.exists():
        # Load "correct" result hashes
        with open(result_hash_library, 'r') as f:
            result_hash_library = json.load(f)
    else:
        result_hash_library = {}

    # Get test names
    baseline_tests = set(baseline.keys())
    result_tests = set(result.keys())

    # Test names must be identical
    diff_set(baseline_tests, result_tests, error='Test names are not identical.')

    item_match_errors = []  # Raise a MatchError for all mismatched values at the end

    for test in baseline_tests:

        # Get baseline and result summary for the specific test
        baseline_summary = baseline[test]
        result_summary = result[test]

        # Swap the baseline and result hashes in the summary
        # for the corresponding hashes in each hash library
        if baseline_hash_library and test in baseline_hash_library and not generating_hashes:
            baseline_summary = replace_hash(baseline_summary, 'baseline_hash',
                                            baseline_hash_library[test])
        if result_hash_library:
            if generating_hashes:  # Newly generate result will appear as baseline_hash
                baseline_summary = replace_hash(baseline_summary, 'baseline_hash',
                                                result_hash_library[test])
            baseline_summary = replace_hash(baseline_summary, 'result_hash',
                                            result_hash_library[test])

        # Get keys of recorded items
        baseline_keys = set(baseline_summary.keys())
        result_keys = set(result_summary.keys())

        # Summaries must have the same keys
        diff_set(baseline_keys, result_keys, error=f'Summary for {test} is not identical.')

        for key in baseline_keys:
            error = f'Summary item {key} for {test} does not match.\n'
            try:
                diff_dict_item(baseline_summary[key], result_summary[key], error=error)
            except MatchError as e:
                item_match_errors.append(str(e))

    if len(item_match_errors) > 0:
        raise MatchError('\n\n----------\n\n'.join(item_match_errors))


def diff_set(baseline, result, error=''):
    """Raise and show the difference between Python sets."""
    if baseline != result:
        missing_from_result = baseline - result
        missing_from_baseline = result - baseline
        if len(missing_from_result) > 0:
            error += f'\nKeys {sorted(missing_from_result)} missing from the result.'
        if len(missing_from_baseline) > 0:
            error += f'\nKeys {sorted(missing_from_baseline)} missing from the baseline.'
        raise MatchError(error)


def diff_dict_item(baseline, result, error=''):
    """Diff a specific item in a pytest-mpl summary dictionary."""
    # Comparison makes the following (good) assumptions
    expected_types = (str, int, float, bool, type(None))
    assert isinstance(baseline, expected_types)
    assert isinstance(result, expected_types)

    # Prepare error message
    error += f'Baseline:\n"{baseline}"\n\n'
    error += f'Result:\n"{result}"\n'

    # Matching items must have the same type
    if type(baseline) is not type(result):
        raise MatchError(error + '\nTypes are not equal.\n')

    # Handle regex in baseline string (so things like paths can be ignored)
    if isinstance(baseline, str) and baseline.startswith('REGEX:'):
        if re.fullmatch(baseline[6:], result) is not None:
            return

    # Handle bool and NoneType
    if isinstance(baseline, (bool, type(None))) and baseline is result:
        return

    # Handle float
    if isinstance(baseline, float) and abs(baseline - result) < 1e-4:
        return

    # Handle str and int
    if baseline == result:
        return

    raise MatchError(error)


def patch_summary(summary, patch_file):
    """Replace in `summary` any items defined in `patch_file`."""
    # By only applying patches, changes between MPL versions are more obvious.
    with open(patch_file, 'r') as f:
        patch = json.load(f)
    for test, test_summary in patch.items():
        for k, v in test_summary.items():
            summary[test][k] = v
    return summary


def replace_hash(summary, hash_key, new_hash):
    """Replace a hash in a pytest-mpl summary with a different hash.

    Parameters
    ----------
    summary : dict
        A single test from a pytest-mpl summary.
    hash_key : str
        Key of the hash. Either `baseline_hash` or `result_hash`.
    new_hash : str
        The new hash.
    """
    assert isinstance(new_hash, str)
    old_hash = summary[hash_key]
    if not isinstance(old_hash, str) or old_hash == new_hash:
        return summary  # Either already correct or missing

    # Update the hash
    summary[hash_key] = new_hash
    summary['status_msg'] = summary['status_msg'].replace(old_hash, new_hash)

    return summary


def assert_existence(summary, items=('baseline_image', 'diff_image', 'result_image'), path=''):
    """Assert that images included in a pytest-mpl summary exist.

    Parameters
    ----------
    summary : dict
        The pytest-mpl summary dictionary to check.
    items : tuple or list, optional
        The image keys to check if reported.
    path : str or path_like, optional, default=''
        Path to results directory. Defaults to current directory.
    """
    for test in summary.values():
        for item in items:
            if test[item] is not None:
                assert (Path(path) / test[item]).exists()


def _escape_regex(msg):
    if not msg.startswith('REGEX:'):
        msg = msg.replace('.', r'\.').replace('(', r'\(').replace(')', r'\)')
        msg = 'REGEX:' + msg
    return msg


def _escape_path(msg, path):
    pattern = (rf"({path}[A-Za-z0-9_\-\/.\\]*)" +
               r"(baseline\\.png|result-failed-diff\\.png|result\\.png|\\.json)")
    msg = re.sub(pattern, r".*\2", msg)
    pattern = rf"({path}[A-Za-z0-9_\-\/.\\]*)"
    msg = re.sub(pattern, r".*", msg)
    return msg


def _escape_float(msg, key):
    pattern = rf"({key}[0-9]+\\\.[0-9]{{1}})([0-9]+)"
    msg = re.sub(pattern, r"\1[0-9]*", msg)
    return msg


def apply_regex(file, regex_paths, regex_strs):
    """Convert all `status_msg` entries in JSON summary file to regex.

    Use in your own script to assist with updating baseline summaries.

    Parameters
    ----------
    file : Path
        JSON summary file to convert `status_msg` to regex in. Overwritten.
    regex_paths : list of str
        List of path beginnings to identify paths that need to be converted to regex.
        E.g. `['/home/user/']`
        Does: `aaa /home/user/pytest/tmp/result\\.png bbb` -> `aaa .*result\\.png bbb`
    regex_strs : list of str
        List of keys to convert following floats to 1 d.p.
        E.g. ['RMS Value: ']
        Does: `aaa RMS Value: 12\\.432644 bbb` -> `aaa RMS Value: 12\\.4[0-9]* bbb`
    """

    with open(file, 'r') as f:
        summary = json.load(f)

    for test in summary.keys():

        msg = summary[test]['status_msg']

        for signal in [*regex_paths, *regex_strs]:
            if signal in msg:
                msg = _escape_regex(msg)
        if not msg.startswith('REGEX:'):
            continue

        for signal in regex_paths:
            if signal in msg:
                msg = _escape_path(msg, path=signal)

        for signal in regex_strs:
            if signal in msg:
                msg = _escape_float(msg, key=signal)

        summary[test]['status_msg'] = msg

    with open(file, 'w') as f:
        json.dump(summary, f, indent=2)


def remove_specific_hashes(summary_file):
    """Replace all hashes in a summary file with placeholder values.

    This is done because the actual hashes used for testing are taken from
    separate files for each specific matplotlib version.
    """

    baseline_placeholder = "###_BASELINE_HASH_###"
    result_placeholder = "###_RESULT_HASH_###"

    with open(summary_file, "r") as f:
        summary = json.load(f)

    for test in summary.keys():

        # Get actual hashes
        baseline = summary[test]["baseline_hash"]
        result = summary[test]["result_hash"]

        # Replace with placeholders (if summary has hashes)
        if baseline is not None:
            summary[test]["baseline_hash"] = baseline_placeholder
            summary[test]["status_msg"] = \
                summary[test]["status_msg"].replace(baseline, baseline_placeholder)
        if result is not None:
            summary[test]["result_hash"] = result_placeholder
            summary[test]["status_msg"] = \
                summary[test]["status_msg"].replace(result, result_placeholder)

    with open(summary_file, "w") as f:
        json.dump(summary, f, indent=2)


def transform_hashes(hash_file):
    """Make hash comparison tests fail correctly.

    Makes hashes of tests *hdiff* in hash_file fail hash comparison
    and remove *hmissing* hashes that should be missing.
    """

    with open(hash_file, "r") as f:
        hashes = json.load(f)

    for test in list(hashes.keys()):
        h = hashes[test]
        if "hdiff" in test and h is not None:
            # Replace first four letters with d1ff to force mismatch
            hashes[test] = "d1ff" + h[4:]
        if "hmissing" in test and h is not None:
            # Remove hashes that should be missing
            del hashes[test]

    with open(hash_file, "w") as f:
        json.dump(hashes, f, indent=2)


def transform_images(baseline_path):
    """Make image comparison tests fail correctly.

    Makes images of tests *idiff* under baseline_path fail image comparison
    and deletes images for *imissing* tests.
    """

    # Delete imissing files
    for file in baseline_path.glob("**/*imissing*.png"):
        file.unlink()

    # Add red cross to idiff files
    for file in baseline_path.glob("**/*idiff*.png"):
        with Image.open(file) as im:
            draw = ImageDraw.Draw(im)
            draw.line((0, 0) + im.size, "#f00", 3)
            draw.line((0, im.size[1], im.size[0], 0), "#f00", 3)
            im.save(file)

    # Resize idiffshape files
    for file in baseline_path.glob("**/*idiffshape*.png"):
        with Image.open(file) as im:
            (width, height) = (im.width // 2, im.height // 2)
            im_resized = im.resize((width, height))
            im_resized.save(file)



import json

import pytest
from helpers import pytester_path


@pytest.mark.parametrize(
    "ini, cli, kwarg, success_expected",
    [
        ("bad", None, None, False),
        ("good", None, None, True),
        ("bad", "good", None, True),
        ("bad", "bad", "good", False),  # Note: CLI overrides kwarg
        ("bad", "good", "bad", True),
    ],
)
def test_config(pytester, ini, cli, kwarg, success_expected):
    path = pytester_path(pytester)
    hash_libraries = {
        "good": path / "good_hash_library.json",
        "bad": path / "bad_hash_library.json",
    }
    ini = f"mpl-hash-library = {hash_libraries[ini]}" if ini else ""
    pytester.makeini(
        f"""
        [pytest]
        mpl-deterministic: true
        {ini}
        """
    )
    kwarg = f"hash_library=r'{hash_libraries[kwarg]}'" if kwarg else ""
    pytester.makepyfile(
        f"""
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare({kwarg})
        def test_mpl():
            fig, ax = plt.subplots()
            ax.plot([1, 3, 2])
            return fig
        """
    )
    pytester.runpytest(f"--mpl-generate-hash-library={hash_libraries['good']}")
    with open(hash_libraries["bad"], "w") as fp:
        json.dump({"test_config.test_mpl": "bad-value"}, fp)
    cli = f"--mpl-hash-library={hash_libraries[cli]}" if cli else ""
    result = pytester.runpytest("--mpl", cli)
    if success_expected:
        result.assert_outcomes(passed=1)
    else:
        result.assert_outcomes(failed=1)



import os
import sys
import json
import shutil
import tempfile
import subprocess
from pathlib import Path

import matplotlib.ft2font
import pytest
from packaging.version import Version

from .helpers import (apply_regex, assert_existence, diff_summary, patch_summary,
                      remove_specific_hashes, transform_hashes, transform_images)

# Handle Matplotlib and FreeType versions
MPL_VERSION = Version(matplotlib.__version__)
FTV = matplotlib.ft2font.__freetype_version__.replace('.', '')
VERSION_ID = f"mpl{MPL_VERSION.major}{MPL_VERSION.minor}_ft{FTV}"
HASH_LIBRARY = Path(__file__).parent / 'subtest' / 'hashes' / (VERSION_ID + ".json")
RESULT_LIBRARY = Path(__file__).parent / 'result_hashes' / (VERSION_ID + ".json")
HASH_LIBRARY_FLAG = rf'--mpl-hash-library={HASH_LIBRARY}'
FULL_BASELINE_PATH = Path(__file__).parent / 'subtest' / 'baseline'

BASELINE_IMAGES_FLAG_REL = ['--mpl-baseline-path=baseline', '--mpl-baseline-relative']
BASELINE_IMAGES_FLAG_ABS = rf'--mpl-baseline-path={FULL_BASELINE_PATH}'

IMAGE_COMPARISON_MODE = ["-k", "image"]
HASH_COMPARISON_MODE = ["-k", "hash"]
# HYBRID_MODE = []

TEST_FILE = Path(__file__).parent / 'subtest'

# Global settings to update baselines when running pytest
# Note: when updating baseline make sure you don't commit "fixes"
# for tests that are expected to fail
# (See also `run_subtest` argument `update_baseline` and `update_summary`.)
UPDATE_BASELINE = os.getenv("MPL_UPDATE_BASELINE") is not None  # baseline images and hashes
UPDATE_SUMMARY = os.getenv("MPL_UPDATE_SUMMARY") is not None  # baseline summaries

# When updating baseline summaries, replace parts of status_msg with regex.
# See helpers.apply_regex for more information.
REGEX_PATHS = [
    str(Path(__file__).parent),  # replace all references to baseline files
    os.path.realpath(tempfile.gettempdir()),  # replace all references to output files
]
REGEX_STRS = [
    r'RMS Value: ',
]


def run_subtest(baseline_summary_name, tmp_path, args, summaries=None, xfail=True,
                has_result_hashes=False, generating_hashes=False,
                update_baseline=UPDATE_BASELINE, update_summary=UPDATE_SUMMARY):
    """ Run pytest (within pytest) and check JSON summary report.

    Parameters
    ----------
    baseline_summary_name : str
        String of the filename without extension for the baseline summary.
    tmp_path : pathlib.Path
        Path of a temporary directory to store results.
    args : list
        Extra arguments to pass to pytest.
    summaries : tuple or list or set, optional, default=[]
        Summaries to generate in addition to `json`.
    xfail : bool, optional, default=True
        Whether the overall pytest run should fail.
    has_result_hashes : bool or str, optional, default=False
        Whether a hash library is expected to exist in the results directory.
        If a string, this is the name of the expected results file.
    generating_hashes : bool, optional, default=False
        Whether `--mpl-generate-hash-library` was specified and
        both of `--mpl-hash-library` and `hash_library=` were not.
    """
    if update_baseline and update_summary:
        raise ValueError("Cannot enable both `update_baseline` and `update_summary`.")

    # Parse arguments
    if summaries is None:
        summaries = []
    assert isinstance(summaries, (tuple, list, set))
    summaries = ','.join({'json'} | set(summaries))

    # Create the results path
    results_path = tmp_path / 'results'
    results_path.mkdir()

    # Configure the arguments to run the test
    pytest_args = [sys.executable, '-m', 'pytest', str(TEST_FILE)]
    mpl_args = ['--mpl', rf'--mpl-results-path={results_path.as_posix()}',
                f'--mpl-generate-summary={summaries}']
    if update_baseline:
        mpl_args += [rf'--mpl-generate-path={FULL_BASELINE_PATH}']
        if HASH_LIBRARY.exists():
            mpl_args += [rf'--mpl-generate-hash-library={HASH_LIBRARY}']

    # Run the test and record exit status
    status = subprocess.call(pytest_args + mpl_args + args)

    # If updating baseline, don't check summaries
    if update_baseline:
        assert status == 0
        transform_images(FULL_BASELINE_PATH)  # Make image comparison tests fail correctly
        if HASH_LIBRARY.exists():
            shutil.copy(HASH_LIBRARY, RESULT_LIBRARY)
            transform_hashes(HASH_LIBRARY)  # Make hash comparison tests fail correctly
        pytest.skip("Skipping testing, since `update_baseline` is enabled.")
        return

    # Ensure exit status is as expected
    if xfail:
        assert status != 0
    else:
        assert status == 0

    # Load summaries
    baseline_path = Path(__file__).parent / 'summaries'
    baseline_file = baseline_path / (baseline_summary_name + '.json')
    results_file = results_path / 'results.json'
    if update_summary:
        shutil.copy(results_file, baseline_file)
        apply_regex(baseline_file, REGEX_PATHS, REGEX_STRS)
        remove_specific_hashes(baseline_file)
    with open(baseline_file, 'r') as f:
        baseline_summary = json.load(f)
    with open(results_file, 'r') as f:
        result_summary = json.load(f)

    # Apply version specific patches
    patch = baseline_path / (baseline_summary_name + f'_{VERSION_ID}.patch.json')
    if patch.exists():
        baseline_summary = patch_summary(baseline_summary, patch)
    # Note: version specific hashes should be handled by diff_summary instead

    # Compare summaries
    diff_summary(baseline_summary, result_summary,
                 baseline_hash_library=HASH_LIBRARY, result_hash_library=RESULT_LIBRARY,
                 generating_hashes=generating_hashes)

    # Ensure reported images exist
    assert_existence(result_summary, path=results_path)

    # Get expected name for the hash library saved to the results directory
    if isinstance(has_result_hashes, str):
        result_hash_file = tmp_path / 'results' / has_result_hashes
        has_result_hashes = True  # convert to bool after processing str
    else:
        result_hash_file = tmp_path / 'results' / HASH_LIBRARY.name

    # Compare the generated hash library to the expected hash library
    if has_result_hashes:
        assert result_hash_file.exists()
        with open(RESULT_LIBRARY, "r") as f:
            baseline = json.load(f)
        with open(result_hash_file, "r") as f:
            result = json.load(f)

        # Baseline contains hashes for all subtests so remove ones not used
        for test in list(baseline.keys()):
            if test not in result:
                del baseline[test]

        diff_summary({'a': baseline}, {'a': result})
    else:
        assert not result_hash_file.exists()

    if update_summary:
        pytest.skip("Skipping testing, since `update_summary` is enabled.")


def test_default(tmp_path):
    run_subtest('test_default', tmp_path, [*IMAGE_COMPARISON_MODE])


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_hash(tmp_path):
    run_subtest('test_hash', tmp_path, [HASH_LIBRARY_FLAG, *HASH_COMPARISON_MODE])


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_results_always(tmp_path):
    run_subtest('test_results_always', tmp_path,
                [HASH_LIBRARY_FLAG, BASELINE_IMAGES_FLAG_ABS, '--mpl-results-always'],
                has_result_hashes=True)


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_html(tmp_path):
    run_subtest('test_results_always', tmp_path,
                [HASH_LIBRARY_FLAG, BASELINE_IMAGES_FLAG_ABS], summaries=['html'],
                has_result_hashes=True)
    assert (tmp_path / 'results' / 'fig_comparison.html').exists()
    assert (tmp_path / 'results' / 'extra.js').exists()
    assert (tmp_path / 'results' / 'styles.css').exists()


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_html_hashes_only(tmp_path):
    run_subtest('test_html_hashes_only', tmp_path,
                [HASH_LIBRARY_FLAG, *HASH_COMPARISON_MODE],
                summaries=['html'], has_result_hashes=True)
    assert (tmp_path / 'results' / 'fig_comparison.html').exists()
    assert (tmp_path / 'results' / 'extra.js').exists()
    assert (tmp_path / 'results' / 'styles.css').exists()


def test_html_images_only(tmp_path):
    run_subtest('test_html_images_only', tmp_path, [*IMAGE_COMPARISON_MODE], summaries=['html'])
    assert (tmp_path / 'results' / 'fig_comparison.html').exists()
    assert (tmp_path / 'results' / 'extra.js').exists()
    assert (tmp_path / 'results' / 'styles.css').exists()


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_basic_html(tmp_path):
    run_subtest('test_results_always', tmp_path,
                [HASH_LIBRARY_FLAG, *BASELINE_IMAGES_FLAG_REL], summaries=['basic-html'],
                has_result_hashes=True)
    assert (tmp_path / 'results' / 'fig_comparison_basic.html').exists()


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_generate(tmp_path):
    # generating hashes and images; no testing
    run_subtest('test_generate', tmp_path,
                [rf'--mpl-generate-path={tmp_path}',
                 rf'--mpl-generate-hash-library={tmp_path / "test_hashes.json"}'],
                xfail=False, generating_hashes=True)


def test_generate_images_only(tmp_path):
    # generating images; no testing
    run_subtest('test_generate_images_only', tmp_path,
                [rf'--mpl-generate-path={tmp_path}', *IMAGE_COMPARISON_MODE], xfail=False)


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_generate_hashes_only(tmp_path):
    # generating hashes; testing images
    run_subtest('test_generate_hashes_only', tmp_path,
                [rf'--mpl-generate-hash-library={tmp_path / "test_hashes.json"}'],
                generating_hashes=True)


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_html_generate(tmp_path):
    # generating hashes and images; no testing
    run_subtest('test_html_generate', tmp_path,
                [rf'--mpl-generate-path={tmp_path}',
                 rf'--mpl-generate-hash-library={tmp_path / "test_hashes.json"}'],
                summaries=['html'], xfail=False, has_result_hashes="test_hashes.json",
                generating_hashes=True)
    assert (tmp_path / 'results' / 'fig_comparison.html').exists()


def test_html_generate_images_only(tmp_path):
    # generating images; no testing
    run_subtest('test_html_generate_images_only', tmp_path,
                [rf'--mpl-generate-path={tmp_path}', *IMAGE_COMPARISON_MODE],
                summaries=['html'], xfail=False)
    assert (tmp_path / 'results' / 'fig_comparison.html').exists()


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_html_generate_hashes_only(tmp_path):
    # generating hashes; testing images
    run_subtest('test_html_generate_hashes_only', tmp_path,
                [rf'--mpl-generate-hash-library={tmp_path / "test_hashes.json"}'],
                summaries=['html'], has_result_hashes="test_hashes.json", generating_hashes=True)
    assert (tmp_path / 'results' / 'fig_comparison.html').exists()


@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_html_run_generate_hashes_only(tmp_path):
    # generating hashes; testing hashes
    run_subtest('test_html_hashes_only', tmp_path,
                [rf'--mpl-generate-hash-library={tmp_path / "test_hashes.json"}',
                 HASH_LIBRARY_FLAG, *HASH_COMPARISON_MODE],
                summaries=['html'], has_result_hashes="test_hashes.json")
    assert (tmp_path / 'results' / 'fig_comparison.html').exists()


# Run a hybrid mode test last so if generating hash libraries, it includes all the hashes.
@pytest.mark.skipif(not HASH_LIBRARY.exists(), reason="No hash library for this mpl version")
def test_hybrid(tmp_path):
    run_subtest('test_hybrid', tmp_path, [HASH_LIBRARY_FLAG, BASELINE_IMAGES_FLAG_ABS])



from helpers import pytester_path

PYFILE = (
    """
    import matplotlib.pyplot as plt
    import pytest
    @pytest.mark.mpl_image_compare
    def test_mpl():
        fig, ax = plt.subplots()
        ax.plot([1, 2, 3])
        return fig
    """
)


def test_generate_baseline_images(pytester):
    pytester.makepyfile(PYFILE)
    baseline_dir = pytester_path(pytester) / "alternative_baseline"
    result = pytester.runpytest(f"--mpl-generate-path={baseline_dir}")
    result.assert_outcomes(skipped=1)
    assert (baseline_dir / "test_mpl.png").exists()


def test_generate_baseline_hashes(pytester):
    pytester.makepyfile(PYFILE)
    path = pytester_path(pytester)
    hash_library = path / "alternative_baseline" / "hash_library_1.json"
    result = pytester.runpytest(
        f"--mpl-generate-hash-library={hash_library}",
        f"--mpl-results-path={path}",
    )
    result.assert_outcomes(failed=1)  # this option enables --mpl
    assert hash_library.exists()
    assert (path / "test_generate_baseline_hashes.test_mpl" / "result.png").exists()


def test_generate_baseline_images_and_hashes(pytester):
    pytester.makepyfile(PYFILE)
    path = pytester_path(pytester)
    baseline_dir = path / "alternative_baseline"
    hash_library = path / "alternative_baseline" / "hash_library_1.json"
    result = pytester.runpytest(
        f"--mpl-generate-path={baseline_dir}",
        f"--mpl-generate-hash-library={hash_library}",
    )
    result.assert_outcomes(skipped=1)
    assert (baseline_dir / "test_mpl.png").exists()
    assert hash_library.exists()



import shutil
from pathlib import Path

import pytest
from helpers import pytester_path


@pytest.mark.parametrize(
    "ini, cli, kwarg, expected_baseline_path, success_expected",
    [
        ("dir1", None, None, "dir1", True),
        ("dir1", "dir2", None, "dir2", True),
        ("dir1", "dir2", "dir3", "dir3", True),
        ("dir1", "dir2", "dir3", "dir2", False),
        (None, None, "dir3", "dir3", True),
    ],
)
def test_config(pytester, ini, cli, kwarg, expected_baseline_path, success_expected):
    path = pytester_path(pytester)
    (path / expected_baseline_path).mkdir()
    shutil.copyfile(  # Test will only pass if baseline is at expected path
        Path(__file__).parent / "baseline" / "2.0.x" / "test_base_style.png",
        path / expected_baseline_path / "test_mpl.png",
    )
    ini = f"mpl-baseline-path = {path / ini}" if ini is not None else ""
    pytester.makeini(
        f"""
        [pytest]
        mpl-default-style = fivethirtyeight
        {ini}
        """
    )
    kwarg = f"baseline_dir=r'{path / kwarg}'" if kwarg else ""
    pytester.makepyfile(
        f"""
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare({kwarg})
        def test_mpl():
            fig, ax = plt.subplots()
            ax.plot([1, 2, 3])
            return fig
        """
    )
    cli = f"--mpl-baseline-path={path / cli}" if cli else ""
    result = pytester.runpytest("--mpl", cli)
    if success_expected:
        result.assert_outcomes(passed=1)
    else:
        result.assert_outcomes(failed=1)



import pytest

from .helpers import plot

# ### Test all permutations of:
# baseline hash: match, diff, or missing
# baseline image: match, diff, or missing


# hash match

@pytest.mark.image
@pytest.mark.hash
@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hmatch_imatch():
    return plot([1, 2, 3, 4])


@pytest.mark.image
@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hmatch_idiff():
    return plot([1, 3, 2, 4])


@pytest.mark.image
@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hmatch_idiffshape():
    return plot([4, 2, 3, 1, 2])


@pytest.mark.image
@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hmatch_imissing():
    return plot([4, 3, 2, 1])


# hash diff

@pytest.mark.hash
@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hdiff_imatch():
    return plot([1, 4, 2, 3])


@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hdiff_idiff():
    return plot([1, 2, 4, 3])


@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hdiff_idiffshape():
    return plot([4, 2, 3, 1, 3])


@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hdiff_imissing():
    return plot([3, 2, 4, 1])


# hash missing

@pytest.mark.hash
@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hmissing_imatch():
    return plot([1, 3, 4, 2])


@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hmissing_idiff():
    return plot([1, 4, 3, 2])


@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hmissing_idiffshape():
    return plot([4, 2, 3, 1, 4])


@pytest.mark.mpl_image_compare(savefig_kwargs={'metadata': {'Software': None}})
def test_hmissing_imissing():
    return plot([2, 4, 3, 1])



# Copyright (c) 2015, Thomas P. Robitaille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# The code below includes code adapted from WCSAxes, which is released
# under a 3-clause BSD license and can be found here:
#
#   https://github.com/astrofrog/wcsaxes

import io
import os
import json
import shutil
import hashlib
import logging
import tempfile
import warnings
import contextlib
from pathlib import Path
from urllib.request import urlopen

import pytest
from packaging.version import Version

from pytest_mpl.summary.html import generate_summary_basic_html, generate_summary_html

DEFAULT_STYLE = "classic"
DEFAULT_TOLERANCE = 2
DEFAULT_BACKEND = "agg"

SUPPORTED_FORMATS = {"html", "json", "basic-html"}

SHAPE_MISMATCH_ERROR = """Error: Image dimensions did not match.
  Expected shape: {expected_shape}
    {expected_path}
  Actual shape: {actual_shape}
    {actual_path}"""

PYTEST_LT_7 = Version(pytest.__version__) < Version("7.0.0")

# The following are the subsets of formats supported by the Matplotlib image
# comparison machinery
RASTER_IMAGE_FORMATS = ['png']
VECTOR_IMAGE_FORMATS = ['eps', 'pdf', 'svg']
ALL_IMAGE_FORMATS = RASTER_IMAGE_FORMATS + VECTOR_IMAGE_FORMATS


def _get_item_dir(item):
    path = Path(item.fspath) if PYTEST_LT_7 else item.path
    return path.parent


def _hash_file(in_stream):
    """
    Hashes an already opened file.
    """
    in_stream.seek(0)
    buf = in_stream.read()
    hasher = hashlib.sha256()
    hasher.update(buf)
    return hasher.hexdigest()


def pathify(path):
    """
    Remove non-path safe characters.
    """
    path = Path(path)
    ext = ''
    if path.suffixes[-1][1:] in ALL_IMAGE_FORMATS:
        ext = path.suffixes[-1]
        path = str(path).split(ext)[0]
    path = str(path)
    path = path.replace('[', '_').replace(']', '_')
    path = path.replace('/', '_')
    if path.endswith('_'):
        path = path[:-1]
    return Path(path + ext)


def generate_test_name(item):
    """
    Generate a unique name for the hash for this test.
    """
    if item.cls is not None:
        name = f"{item.module.__name__}.{item.cls.__name__}.{item.name}"
    else:
        name = f"{item.module.__name__}.{item.name}"
    return name


def wrap_figure_interceptor(plugin, item):
    """
    Intercept and store figures returned by test functions.
    """
    # Only intercept figures on marked figure tests
    if get_compare(item) is not None:

        # Use the full test name as a key to ensure correct figure is being retrieved
        test_name = generate_test_name(item)

        def figure_interceptor(store, obj):
            def wrapper(*args, **kwargs):
                store.return_value[test_name] = obj(*args, **kwargs)
            return wrapper

        item.obj = figure_interceptor(plugin, item.obj)


def pytest_report_header():
    import matplotlib
    import matplotlib.ft2font
    return ["Matplotlib: {0}".format(matplotlib.__version__),
            "Freetype: {0}".format(matplotlib.ft2font.__freetype_version__)]


def pytest_addoption(parser):
    group = parser.getgroup("matplotlib image comparison")

    msg = "Enable comparison of matplotlib figures to reference files"
    group.addoption("--mpl", help=msg, action="store_true")

    msg = "directory to generate reference images in, relative to location where py.test is run"
    group.addoption("--mpl-generate-path", help=msg, action="store")

    msg = "filepath to save a generated hash library, relative to location where py.test is run"
    group.addoption("--mpl-generate-hash-library", help=msg, action="store")

    msg = (
        "directory containing baseline images, relative to "
        "location where py.test is run unless --mpl-baseline-relative is given. "
        "This can also be a URL or a set of comma-separated URLs (in case "
        "mirrors are specified)"
    )
    option = "mpl-baseline-path"
    group.addoption(f"--{option}", help=msg, action="store")
    parser.addini(option, help=msg)

    msg = "interpret the baseline directory as relative to the test location."
    group.addoption("--mpl-baseline-relative", help=msg, action="store_true")

    msg = "json library of image hashes, relative to location where py.test is run"
    option = "mpl-hash-library"
    group.addoption(f"--{option}", help=msg, action="store")
    parser.addini(option, help=msg)

    msg = (
        "Generate a summary report of any failed tests"
        ", in --mpl-results-path. The type of the report should be "
        "specified. Supported types are `html`, `json` and `basic-html`. "
        "Multiple types can be specified separated by commas."
    )
    option = "mpl-generate-summary"
    group.addoption(f"--{option}", help=msg, action="store")
    parser.addini(option, help=msg)

    msg = "directory for test results, relative to location where py.test is run"
    option = "mpl-results-path"
    group.addoption(f"--{option}", help=msg, action="store")
    parser.addini(option, help=msg)

    msg = (
        "Always compare to baseline images and save result images, even for passing tests. "
        "This option is automatically applied when generating a HTML summary."
    )
    option = "mpl-results-always"
    group.addoption(f"--{option}", help=msg, action="store_true")
    parser.addini(option, help=msg)

    msg = "use fully qualified test name as the filename."
    option = "mpl-use-full-test-name"
    group.addoption(f"--{option}", help=msg, action="store_true")
    parser.addini(option, help=msg, type="bool")

    msg = "default style to use for tests, unless specified in the mpl_image_compare decorator"
    option = "mpl-default-style"
    group.addoption(f"--{option}", help=msg, action="store")
    parser.addini(option, help=msg)

    msg = "default tolerance to use for tests, unless specified in the mpl_image_compare decorator"
    option = "mpl-default-tolerance"
    group.addoption(f"--{option}", help=msg, action="store")
    parser.addini(option, help=msg)

    msg = "whether to make the image file metadata deterministic"
    option_true = "mpl-deterministic"
    option_false = "mpl-no-deterministic"
    group.addoption(f"--{option_true}", help=msg, action="store_true")
    group.addoption(f"--{option_false}", help=msg, action="store_true")
    parser.addini(option_true, help=msg, type="bool", default=None)

    msg = "default backend to use for tests, unless specified in the mpl_image_compare decorator"
    option = "mpl-default-backend"
    group.addoption(f"--{option}", help=msg, action="store")
    parser.addini(option, help=msg)


def pytest_configure(config):

    config.addinivalue_line(
        "markers",
        "mpl_image_compare: Compares matplotlib figures against a baseline image",
    )

    if (
        config.getoption("--mpl")
        or config.getoption("--mpl-generate-path") is not None
        or config.getoption("--mpl-generate-hash-library") is not None
    ):

        def get_cli_or_ini(name, default=None):
            return config.getoption(f"--{name}") or config.getini(name) or default

        generate_dir = config.getoption("--mpl-generate-path")
        generate_hash_lib = config.getoption("--mpl-generate-hash-library")

        baseline_dir = get_cli_or_ini("mpl-baseline-path")
        if config.getoption("--mpl-baseline-relative"):
            baseline_relative_dir = config.getoption("--mpl-baseline-path")
        else:
            baseline_relative_dir = None
        use_full_test_name = get_cli_or_ini("mpl-use-full-test-name")

        hash_library = get_cli_or_ini("mpl-hash-library")
        _hash_library_from_cli = bool(config.getoption("--mpl-hash-library"))  # for backwards compatibility

        default_tolerance = get_cli_or_ini("mpl-default-tolerance", DEFAULT_TOLERANCE)
        if isinstance(default_tolerance, str):
            if default_tolerance.isdigit():  # prefer int if possible
                default_tolerance = int(default_tolerance)
            else:
                default_tolerance = float(default_tolerance)

        deterministic_ini = config.getini("mpl-deterministic")
        deterministic_flag_true = config.getoption("--mpl-deterministic")
        deterministic_flag_false = config.getoption("--mpl-no-deterministic")
        if deterministic_flag_true and deterministic_flag_false:
            raise ValueError("Only one of `--mpl-deterministic` and `--mpl-no-deterministic` can be set.")
        if deterministic_flag_true:
            deterministic = True
        elif deterministic_flag_false:
            deterministic = False
        elif isinstance(deterministic_ini, bool):
            deterministic = deterministic_ini
        else:
            deterministic = None

        default_style = get_cli_or_ini("mpl-default-style", DEFAULT_STYLE)
        default_backend = get_cli_or_ini("mpl-default-backend", DEFAULT_BACKEND)

        results_dir = get_cli_or_ini("mpl-results-path")
        results_always = get_cli_or_ini("mpl-results-always")
        generate_summary = get_cli_or_ini("mpl-generate-summary")

        if generate_dir is not None:
            if baseline_dir is not None:
                warnings.warn("Ignoring --mpl-baseline-path since --mpl-generate-path is set")

        if baseline_dir is not None and not baseline_dir.startswith(("https", "http")):
            baseline_dir = os.path.abspath(baseline_dir)
        if generate_dir is not None:
            baseline_dir = os.path.abspath(generate_dir)
        if results_dir is not None:
            results_dir = os.path.abspath(results_dir)
        if hash_library is not None:
            # For backwards compatibility, don't make absolute if set via CLI option
            if not _hash_library_from_cli:
                hash_library = os.path.abspath(hash_library)

        plugin = ImageComparison(
            config,
            baseline_dir=baseline_dir,
            baseline_relative_dir=baseline_relative_dir,
            generate_dir=generate_dir,
            results_dir=results_dir,
            hash_library=hash_library,
            generate_hash_library=generate_hash_lib,
            generate_summary=generate_summary,
            results_always=results_always,
            use_full_test_name=use_full_test_name,
            default_style=default_style,
            default_tolerance=default_tolerance,
            deterministic=deterministic,
            default_backend=default_backend,
            _hash_library_from_cli=_hash_library_from_cli,
        )
        config.pluginmanager.register(plugin)

    else:
        config.pluginmanager.register(FigureCloser(config))


@contextlib.contextmanager
def switch_backend(backend):
    import matplotlib
    import matplotlib.pyplot as plt
    prev_backend = matplotlib.get_backend().lower()
    if prev_backend != backend.lower():
        plt.switch_backend(backend)
        yield
        plt.switch_backend(prev_backend)
    else:
        yield


def close_mpl_figure(fig):
    "Close a given matplotlib Figure. Any other type of figure is ignored"

    import matplotlib.pyplot as plt
    from matplotlib.figure import Figure

    # We only need to close actual Matplotlib figure objects. If
    # we are dealing with a figure-like object that provides
    # savefig but is not a real Matplotlib object, we shouldn't
    # try closing it here.
    if isinstance(fig, Figure):
        plt.close(fig)


def get_compare(item):
    """
    Return the mpl_image_compare marker for the given item.
    """
    return item.get_closest_marker("mpl_image_compare")


def path_is_not_none(apath):
    return Path(apath) if apath is not None else apath


class ImageComparison:
    def __init__(
        self,
        config,
        baseline_dir=None,
        baseline_relative_dir=None,
        generate_dir=None,
        results_dir=None,
        hash_library=None,
        generate_hash_library=None,
        generate_summary=None,
        results_always=False,
        use_full_test_name=False,
        default_style=DEFAULT_STYLE,
        default_tolerance=DEFAULT_TOLERANCE,
        deterministic=None,
        default_backend=DEFAULT_BACKEND,
        _hash_library_from_cli=False,  # for backwards compatibility
    ):
        self.config = config
        self.baseline_dir = baseline_dir
        self.baseline_relative_dir = path_is_not_none(baseline_relative_dir)
        self.generate_dir = path_is_not_none(generate_dir)
        self.results_dir = path_is_not_none(results_dir)
        self.hash_library = path_is_not_none(hash_library)
        self._hash_library_from_cli = _hash_library_from_cli  # for backwards compatibility
        self.generate_hash_library = path_is_not_none(generate_hash_library)
        if generate_summary:
            generate_summary = {i.lower() for i in generate_summary.split(',')}
            unsupported_formats = generate_summary - SUPPORTED_FORMATS
            if len(unsupported_formats) > 0:
                raise ValueError(f"The mpl summary type(s) '{sorted(unsupported_formats)}' "
                                 "are not supported.")
            # When generating HTML always apply `results_always`
            if generate_summary & {'html', 'basic-html'}:
                results_always = True
        self.generate_summary = generate_summary
        self.results_always = results_always
        self.use_full_test_name = use_full_test_name

        self.default_style = default_style
        self.default_tolerance = default_tolerance
        self.deterministic = deterministic
        self.default_backend = default_backend

        # Generate the containing dir for all test results
        if not self.results_dir:
            self.results_dir = Path(tempfile.mkdtemp(dir=self.results_dir))
        self.results_dir.mkdir(parents=True, exist_ok=True)

        # Decide what to call the downloadable results hash library
        if self.hash_library is not None:
            self.results_hash_library_name = self.hash_library.name
        else:  # Use the first filename encountered in a `hash_library=` kwarg
            self.results_hash_library_name = None

        # We need global state to store all the hashes generated over the run
        self._generated_hash_library = {}
        self._test_results = {}
        self._test_stats = None
        self.return_value = {}

        # configure a separate logger for this pluggin which is independent
        # of the options that are configured for pytest or for the code that
        # is tested; turn debug prints on only if "-vv" or more passed
        level = logging.DEBUG if config.option.verbose > 1 else logging.INFO
        if config.option.log_cli_format is not None:
            fmt = config.option.log_cli_format
        else:
            # use pytest's default fmt
            fmt = "%(levelname)-8s %(name)s:%(filename)s:%(lineno)d %(message)s"
        formatter = logging.Formatter(fmt)
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        self.logger = logging.getLogger('pytest-mpl')
        self.logger.propagate = False
        self.logger.setLevel(level)
        self.logger.addHandler(handler)

    def _file_extension(self, item):
        compare = get_compare(item)
        savefig_kwargs = compare.kwargs.get('savefig_kwargs', {})
        return savefig_kwargs.get('format', 'png')

    def generate_filename(self, item):
        """
        Given a pytest item, generate the figure filename.
        """
        ext = self._file_extension(item)
        if self.use_full_test_name:
            filename = generate_test_name(item) + f'.{ext}'
        else:
            compare = get_compare(item)
            # Find test name to use as plot name
            filename = compare.kwargs.get('filename', None)
            if filename is None:
                filename = item.name + f'.{ext}'

        filename = str(pathify(filename))
        return filename

    def make_test_results_dir(self, item):
        """
        Generate the directory to put the results in.
        """
        test_name = pathify(generate_test_name(item))
        results_dir = self.results_dir / test_name
        results_dir.mkdir(exist_ok=True, parents=True)
        return results_dir

    def baseline_directory_specified(self, item):
        """
        Returns `True` if a non-default baseline directory is specified.
        """
        compare = get_compare(item)
        item_baseline_dir = compare.kwargs.get('baseline_dir', None)
        return item_baseline_dir or self.baseline_dir or self.baseline_relative_dir

    def get_baseline_directory(self, item):
        """
        Return a full path to the baseline directory, either local or remote.

        Using the global and per-test configuration return the absolute
        baseline dir, if the baseline file is local else return base URL.
        """
        compare = get_compare(item)
        baseline_dir = compare.kwargs.get('baseline_dir', None)
        if baseline_dir is None:
            if self.baseline_dir is None:
                baseline_dir = _get_item_dir(item) / 'baseline'
            else:
                if self.baseline_relative_dir:
                    # baseline dir is relative to the current test
                    baseline_dir = _get_item_dir(item) / self.baseline_relative_dir
                else:
                    # baseline dir is relative to where pytest was run
                    baseline_dir = self.baseline_dir

        baseline_remote = (isinstance(baseline_dir, str) and  # noqa
                           baseline_dir.startswith(('http://', 'https://')))
        if not baseline_remote:
            return _get_item_dir(item) / baseline_dir

        return baseline_dir

    def _download_file(self, baseline, filename):
        # Note that baseline can be a comma-separated list of URLs that we can
        # then treat as mirrors
        for base_url in baseline.split(','):
            try:
                u = urlopen(base_url + filename)
                content = u.read()
            except Exception as e:
                self.logger.info(f'Downloading {base_url + filename} failed: {repr(e)}')
            else:
                break
        else:  # Could not download baseline image from any of the available URLs
            return
        result_dir = Path(tempfile.mkdtemp())
        filename = result_dir / 'downloaded'
        with open(str(filename), 'wb') as tmpfile:
            tmpfile.write(content)
        return Path(filename)

    def obtain_baseline_image(self, item):
        """
        Copy the baseline image to our working directory.

        If the image is remote it is downloaded, if it is local it is copied to
        ensure it is kept in the event of a test failure.
        """
        filename = self.generate_filename(item)
        baseline_dir = self.get_baseline_directory(item)
        baseline_remote = (isinstance(baseline_dir, str) and  # noqa
                           baseline_dir.startswith(('http://', 'https://')))
        if baseline_remote:
            # baseline_dir can be a list of URLs when remote, so we have to
            # pass base and filename to download
            baseline_image = self._download_file(baseline_dir, filename)
        else:
            baseline_image = (baseline_dir / filename).absolute()

        return baseline_image

    def generate_baseline_image(self, item, fig):
        """
        Generate reference figures.
        """

        if not os.path.exists(self.generate_dir):
            os.makedirs(self.generate_dir)

        baseline_filename = self.generate_filename(item)
        baseline_path = (self.generate_dir / baseline_filename).absolute()
        self.save_figure(item, fig, baseline_path)
        close_mpl_figure(fig)

        return baseline_path

    def generate_image_hash(self, item, fig):
        """
        For a `matplotlib.figure.Figure`, returns the SHA256 hash as a hexadecimal
        string.
        """

        imgdata = io.BytesIO()
        self.save_figure(item, fig, imgdata)
        out = _hash_file(imgdata)
        imgdata.close()

        close_mpl_figure(fig)
        return out

    def compare_image_to_baseline(self, item, fig, result_dir, summary=None):
        """
        Compare a test image to a baseline image.
        """
        from matplotlib.image import imread
        from matplotlib.testing.compare import compare_images

        if summary is None:
            summary = {}

        compare = get_compare(item)
        tolerance = compare.kwargs.get('tolerance', self.default_tolerance)

        ext = self._file_extension(item)

        test_image = (result_dir / f"result.{ext}").absolute()
        self.save_figure(item, fig, test_image)

        if ext in ['png', 'svg']:  # Use original file
            summary['result_image'] = test_image.relative_to(self.results_dir).as_posix()
        else:
            summary['result_image'] = (result_dir / f"result_{ext}.png").relative_to(self.results_dir).as_posix()

        baseline_image_ref = self.obtain_baseline_image(item)

        baseline_missing = None
        if baseline_image_ref is None:
            baseline_missing = ("Could not download the baseline image from "
                                "any of the available URLs.\n")
        elif not os.path.exists(baseline_image_ref):
            baseline_missing = ("Image file not found for comparison test in: \n\t"
                                f"{self.get_baseline_directory(item)}\n")

        if baseline_missing:
            summary['status'] = 'failed'
            summary['image_status'] = 'missing'
            error_message = (baseline_missing +
                             "(This is expected for new tests.)\n"
                             "Generated Image: \n\t"
                             f"{test_image}")
            summary['status_msg'] = error_message
            return error_message

        # setuptools may put the baseline images in non-accessible places,
        # copy to our tmpdir to be sure to keep them in case of failure
        baseline_image = (result_dir / f"baseline.{ext}").absolute()
        shutil.copyfile(baseline_image_ref, baseline_image)

        if ext in ['png', 'svg']:  # Use original file
            summary['baseline_image'] = baseline_image.relative_to(self.results_dir).as_posix()
        else:
            summary['baseline_image'] = (result_dir / f"baseline_{ext}.png").relative_to(self.results_dir).as_posix()

        # Compare image size ourselves since the Matplotlib
        # exception is a bit cryptic in this case and doesn't show
        # the filenames. However imread won't work for vector graphics so we
        # only do this for raster files.
        if ext in RASTER_IMAGE_FORMATS:
            expected_shape = imread(str(baseline_image)).shape[:2]
            actual_shape = imread(str(test_image)).shape[:2]
            if expected_shape != actual_shape:
                summary['status'] = 'failed'
                summary['image_status'] = 'diff'
                error_message = SHAPE_MISMATCH_ERROR.format(expected_path=baseline_image,
                                                            expected_shape=expected_shape,
                                                            actual_path=test_image,
                                                            actual_shape=actual_shape)
                summary['status_msg'] = error_message
                return error_message

        results = compare_images(str(baseline_image), str(test_image), tol=tolerance, in_decorator=True)

        summary['tolerance'] = tolerance
        if results is None:
            summary['status'] = 'passed'
            summary['image_status'] = 'match'
            summary['status_msg'] = 'Image comparison passed.'
            return None
        else:
            summary['status'] = 'failed'
            summary['image_status'] = 'diff'
            summary['rms'] = results['rms']
            summary['diff_image'] = Path(results['diff']).relative_to(self.results_dir).as_posix()
            template = ['Error: Image files did not match.',
                        'RMS Value: {rms}',
                        'Expected:  \n    {expected}',
                        'Actual:    \n    {actual}',
                        'Difference:\n    {diff}',
                        'Tolerance: \n    {tol}', ]
            error_message = '\n  '.join([line.format(**results) for line in template])
            summary['status_msg'] = error_message
            return error_message

    def load_hash_library(self, library_path):
        with open(str(library_path)) as fp:
            return json.load(fp)

    def save_figure(self, item, fig, filename):
        if isinstance(filename, Path):
            filename = str(filename)
        compare = get_compare(item)
        savefig_kwargs = compare.kwargs.get('savefig_kwargs', {})
        deterministic = compare.kwargs.get('deterministic', self.deterministic)

        original_source_date_epoch = os.environ.get('SOURCE_DATE_EPOCH', None)

        extra_rcparams = {}

        ext = self._file_extension(item)

        if deterministic is None:

            # The deterministic option should only matter for hash-based tests,
            # so we first check if a hash library is being used

            if self.hash_library or compare.kwargs.get('hash_library', None):

                if ext == 'png':
                    if 'metadata' not in savefig_kwargs or 'Software' not in savefig_kwargs['metadata']:
                        warnings.warn("deterministic option not set (currently defaulting to False), "
                                      "in future this will default to True to give consistent "
                                      "hashes across Matplotlib versions. To suppress this warning, "
                                      "set deterministic to True if you are happy with the future "
                                      "behavior or to False if you want to preserve the old behavior.",
                                      FutureWarning)
                    else:
                        # Set to False but in practice because Software is set to a constant value
                        # by the caller, the output will be deterministic (we don't want to change
                        # Software to None if the caller set it to e.g. 'test')
                        deterministic = False
                else:
                    deterministic = True

            else:

                # We can just default to True since it shouldn't matter and in
                # case generated images are somehow used in future to compute
                # hashes

                deterministic = True

        if deterministic:

            # Make sure we don't modify the original dictionary in case is a common
            # object used by different tests
            savefig_kwargs = savefig_kwargs.copy()

            if 'metadata' not in savefig_kwargs:
                savefig_kwargs['metadata'] = {}

            if ext == 'png':
                extra_metadata = {"Software": None}
            elif ext == 'pdf':
                extra_metadata = {"Creator": None, "Producer": None, "CreationDate": None}
            elif ext == 'eps':
                extra_metadata = {"Creator": "test"}
                os.environ['SOURCE_DATE_EPOCH'] = '1680254601'
            elif ext == 'svg':
                extra_metadata = {"Date": None}
                extra_rcparams["svg.hashsalt"] = "test"

            savefig_kwargs['metadata'].update(extra_metadata)

        import matplotlib.pyplot as plt

        with plt.rc_context(rc=extra_rcparams):
            fig.savefig(filename, **savefig_kwargs)

        if original_source_date_epoch is not None:
            os.environ['SOURCE_DATE_EPOCH'] = original_source_date_epoch

    def compare_image_to_hash_library(self, item, fig, result_dir, summary=None):
        hash_comparison_pass = False
        if summary is None:
            summary = {}

        compare = get_compare(item)

        ext = self._file_extension(item)

        if not self.results_hash_library_name:
            # Use hash library name of current test as results hash library name
            self.results_hash_library_name = Path(compare.kwargs.get("hash_library", "")).name

        # Order of precedence for hash library: CLI, kwargs, INI (for backwards compatibility)
        hash_library_filename = compare.kwargs.get("hash_library", None) or self.hash_library
        if self._hash_library_from_cli:  # for backwards compatibility
            hash_library_filename = self.hash_library
        hash_library_filename = _get_item_dir(item) / hash_library_filename

        if not Path(hash_library_filename).exists():
            pytest.fail(f"Can't find hash library at path {hash_library_filename}")

        hash_library = self.load_hash_library(hash_library_filename)
        hash_name = generate_test_name(item)
        baseline_hash = hash_library.get(hash_name, None)
        summary['baseline_hash'] = baseline_hash

        test_hash = self.generate_image_hash(item, fig)
        summary['result_hash'] = test_hash

        if baseline_hash is None:  # hash-missing
            summary['status'] = 'failed'
            summary['hash_status'] = 'missing'
            summary['status_msg'] = (f"Hash for test '{hash_name}' not found in {hash_library_filename}. "
                                     f"Generated hash is {test_hash}.")
        elif test_hash == baseline_hash:  # hash-match
            hash_comparison_pass = True
            summary['status'] = 'passed'
            summary['hash_status'] = 'match'
            summary['status_msg'] = 'Test hash matches baseline hash.'
        else:  # hash-diff
            summary['status'] = 'failed'
            summary['hash_status'] = 'diff'
            summary['status_msg'] = (f"Hash {test_hash} doesn't match hash "
                                     f"{baseline_hash} in library "
                                     f"{hash_library_filename} for test {hash_name}.")

        # Save the figure for later summary (will be removed later if not needed)
        test_image = (result_dir / f"result.{ext}").absolute()
        self.save_figure(item, fig, test_image)
        summary['result_image'] = test_image.relative_to(self.results_dir).as_posix()

        # Hybrid mode (hash and image comparison)
        if self.baseline_directory_specified(item):

            # Skip image comparison if hash matches (unless `--mpl-results-always`)
            if hash_comparison_pass and not self.results_always:
                return

            # Run image comparison
            baseline_summary = {}  # summary for image comparison to merge with hash comparison summary
            try:  # Ignore all errors as success does not influence the overall test result
                baseline_comparison = self.compare_image_to_baseline(item, fig, result_dir,
                                                                     summary=baseline_summary)
            except Exception as baseline_error:  # Append to test error later
                summary['image_status'] = 'diff'  # (not necessarily diff, but makes user aware)
                baseline_comparison = str(baseline_error)
            else:  # Update main summary
                for k in ['image_status', 'baseline_image', 'diff_image',
                          'rms', 'tolerance', 'result_image']:
                    summary[k] = summary[k] or baseline_summary.get(k)

            # Append the log from image comparison
            r = baseline_comparison or "The comparison to the baseline image succeeded."
            summary['status_msg'] += ("\n\n"
                                      "Image comparison test\n"
                                      "---------------------\n") + r

        if hash_comparison_pass:  # Return None to indicate test passed
            return
        return summary['status_msg']

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_call(self, item):  # noqa

        compare = get_compare(item)

        if compare is None:
            yield
            return

        import matplotlib.pyplot as plt
        try:
            from matplotlib.testing.decorators import remove_ticks_and_titles
        except ImportError:
            from matplotlib.testing.decorators import ImageComparisonTest as MplImageComparisonTest
            remove_ticks_and_titles = MplImageComparisonTest.remove_text

        style = compare.kwargs.get('style', self.default_style)
        remove_text = compare.kwargs.get('remove_text', False)
        backend = compare.kwargs.get('backend', self.default_backend)

        ext = self._file_extension(item)

        with plt.style.context(style, after_reset=True), switch_backend(backend):

            test_name = generate_test_name(item)

            # Store fallback summary in case of exceptions
            summary = {
                'status': 'failed',
                'image_status': None,
                'hash_status': None,
                'status_msg': 'An exception was raised while testing the figure.',
                'baseline_image': None,
                'diff_image': None,
                'rms': None,
                'tolerance': None,
                'result_image': None,
                'baseline_hash': None,
                'result_hash': None,
            }
            self._test_results[test_name] = summary

            # Run test and get figure object
            wrap_figure_interceptor(self, item)

            # See https://github.com/pytest-dev/pytest/issues/11714
            result = yield
            try:
                if test_name not in self.return_value:
                    # Test function did not complete successfully
                    summary['status'] = 'failed'
                    summary['status_msg'] = ('Test function raised an exception '
                                             'before returning a figure.')
                    self._test_results[test_name] = summary
                    return
                fig = self.return_value[test_name]

                if remove_text:
                    remove_ticks_and_titles(fig)

                result_dir = self.make_test_results_dir(item)

                # What we do now depends on whether we are generating the
                # reference images or simply running the test.
                if self.generate_dir is not None:
                    summary['status'] = 'skipped'
                    summary['image_status'] = 'generated'
                    summary['status_msg'] = 'Skipped test, since generating image.'
                    generate_image = self.generate_baseline_image(item, fig)
                    if self.results_always:  # Make baseline image available in HTML
                        result_image = (result_dir / f"baseline.{ext}").absolute()
                        shutil.copy(generate_image, result_image)
                        summary['baseline_image'] = \
                            result_image.relative_to(self.results_dir).as_posix()

                if self.generate_hash_library is not None:
                    summary['hash_status'] = 'generated'
                    image_hash = self.generate_image_hash(item, fig)
                    self._generated_hash_library[test_name] = image_hash
                    summary['baseline_hash'] = image_hash

                # Only test figures if not generating images
                if self.generate_dir is None:
                    # Compare to hash library
                    if self.hash_library or compare.kwargs.get('hash_library', None):
                        msg = self.compare_image_to_hash_library(item, fig, result_dir, summary=summary)

                    # Compare against a baseline if specified
                    else:
                        msg = self.compare_image_to_baseline(item, fig, result_dir, summary=summary)

                    close_mpl_figure(fig)

                    if msg is None:
                        if not self.results_always:
                            shutil.rmtree(result_dir)
                            for image_type in ['baseline_image', 'diff_image', 'result_image']:
                                summary[image_type] = None  # image no longer exists
                    else:
                        self._test_results[test_name] = summary
                        pytest.fail(msg, pytrace=False)

                close_mpl_figure(fig)

                self._test_results[test_name] = summary

                if summary['status'] == 'skipped':
                    pytest.skip(summary['status_msg'])
            except BaseException as e:
                if hasattr(result, "force_exception"):  # pluggy>=1.2.0
                    result.force_exception(e)
                else:
                    result._result = None
                    result._excinfo = (type(e), e, e.__traceback__)

    def generate_summary_json(self):
        json_file = self.results_dir / 'results.json'
        with open(json_file, 'w') as f:
            json.dump(self._test_results, f, indent=2)
        return json_file

    def pytest_unconfigure(self, config):
        """
        Save out the hash library at the end of the run.
        """
        result_hash_library = self.results_dir / (self.results_hash_library_name or "temp.json")
        if self.generate_hash_library is not None:
            hash_library_path = Path(config.rootdir) / self.generate_hash_library
            hash_library_path.parent.mkdir(parents=True, exist_ok=True)
            with open(hash_library_path, "w") as fp:
                json.dump(self._generated_hash_library, fp, indent=2)
            if self.results_always:  # Make accessible in results directory
                # Use same name as generated
                result_hash_library = self.results_dir / hash_library_path.name
                shutil.copy(hash_library_path, result_hash_library)
        elif self.results_always and self.results_hash_library_name:
            result_hashes = {k: v['result_hash'] for k, v in self._test_results.items()
                             if v['result_hash']}
            if len(result_hashes) > 0:  # At least one hash comparison test
                with open(result_hash_library, "w") as fp:
                    json.dump(result_hashes, fp, indent=2)

        if self.generate_summary:
            kwargs = {}
            if 'json' in self.generate_summary:
                summary = self.generate_summary_json()
                print(f"A JSON report can be found at: {summary}")
            if result_hash_library.exists():  # link to it in the HTML
                kwargs["hash_library"] = result_hash_library.name
            if 'html' in self.generate_summary:
                summary = generate_summary_html(self._test_results, self.results_dir, **kwargs)
                print(f"A summary of test results can be found at: {summary}")
            if 'basic-html' in self.generate_summary:
                summary = generate_summary_basic_html(self._test_results, self.results_dir,
                                                      **kwargs)
                print(f"A summary of test results can be found at: {summary}")


class FigureCloser:
    """
    This is used in place of ImageComparison when the --mpl option is not used,
    to make sure that we still close figures returned by tests.
    """

    def __init__(self, config):
        self.config = config
        self.return_value = {}

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_call(self, item):
        wrap_figure_interceptor(self, item)
        yield
        if get_compare(item) is not None:
            test_name = generate_test_name(item)
            if test_name not in self.return_value:
                # Test function did not complete successfully
                return
            fig = self.return_value[test_name]
            close_mpl_figure(fig)



import os
import sys
import json
import subprocess
from pathlib import Path
from unittest import TestCase

import matplotlib
import matplotlib.ft2font
import matplotlib.pyplot as plt
import pytest
from helpers import skip_if_format_unsupported
from packaging.version import Version

MPL_VERSION = Version(matplotlib.__version__)

baseline_dir = 'baseline'

if MPL_VERSION >= Version('2'):
    baseline_subdir = '2.0.x'

baseline_dir_local = os.path.join(baseline_dir, baseline_subdir)
baseline_dir_remote = 'http://matplotlib.github.io/pytest-mpl/' + baseline_subdir + '/'

ftv = matplotlib.ft2font.__freetype_version__.replace('.', '')
hash_filename = f"mpl{MPL_VERSION.major}{MPL_VERSION.minor}_ft{ftv}.json"

if "+" in matplotlib.__version__:
    hash_filename = "mpldev.json"

hash_library = (Path(__file__).parent / "baseline" /  # noqa
                "hashes" / hash_filename)

fail_hash_library = Path(__file__).parent / "baseline" / "test_hash_lib.json"
baseline_dir_abs = Path(__file__).parent / "baseline" / baseline_subdir
hash_baseline_dir_abs = Path(__file__).parent / "baseline" / "hybrid"


WIN = sys.platform.startswith('win')

# In some cases, the fonts on Windows can be quite different
DEFAULT_TOLERANCE = 10 if WIN else 2


def call_pytest(args):
    return subprocess.call([sys.executable, '-m', 'pytest', '-s'] + args)


def assert_pytest_fails_with(args, output_substring):
    try:
        subprocess.check_output([sys.executable, '-m', 'pytest', '-s'] + args)
    except subprocess.CalledProcessError as exc:
        output = exc.output.decode()
        assert output_substring in output, output
        return output
    else:
        raise RuntimeError(f'pytest did not fail with args {args}')


@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_local,
                               tolerance=DEFAULT_TOLERANCE)
def test_succeeds():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 3])
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_remote,
                               tolerance=DEFAULT_TOLERANCE)
def test_succeeds_remote():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 3])
    return fig


# The following tries an invalid URL first (or at least a URL where the baseline
# image won't exist), but should succeed with the second mirror.
@pytest.mark.mpl_image_compare(baseline_dir='http://www.python.org,' + baseline_dir_remote,
                               filename='test_succeeds_remote.png',
                               tolerance=DEFAULT_TOLERANCE)
def test_succeeds_faulty_mirror():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 3])
    return fig


class TestClass(object):

    @pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_local,
                                   tolerance=DEFAULT_TOLERANCE)
    def test_succeeds(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot([1, 2, 3])
        return fig


@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_local,
                               savefig_kwargs={'dpi': 30},
                               tolerance=DEFAULT_TOLERANCE)
def test_dpi():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 3])
    return fig


TEST_FAILING = """
import pytest
import matplotlib.pyplot as plt
@pytest.mark.mpl_image_compare
def test_fail():
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot([1,2,2])
    return fig
"""


def test_fails(tmp_path):
    test_file = tmp_path / "test.py"
    test_file.write_text(TEST_FAILING)
    test_file = str(test_file)

    # If we use --mpl, it should detect that the figure is wrong
    code = call_pytest(['--mpl', test_file])
    assert code != 0

    # If we don't use --mpl option, the test should succeed
    code = call_pytest([test_file])
    assert code == 0


TEST_OUTPUT_DIR = """
import pytest
import matplotlib.pyplot as plt
@pytest.mark.mpl_image_compare
def test_output_dir():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 3])
    return fig
"""


def test_output_dir(tmp_path):
    test_file = tmp_path / "test.py"
    test_file.write_text(TEST_OUTPUT_DIR)

    output_dir = tmp_path / "test_output_dir"

    # When we run the test, we should get output images where we specify
    code = call_pytest([f'--mpl-results-path={output_dir}',
                        '--mpl', str(test_file)])

    assert code != 0
    assert output_dir.exists()
    assert (output_dir / "test.test_output_dir" / "result.png").exists()


TEST_GENERATE = """
import pytest
import matplotlib.pyplot as plt
@pytest.mark.mpl_image_compare
def test_gen():
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot([1,2,3])
    return fig
"""


def test_generate(tmp_path):

    test_file = tmp_path / "test.py"
    test_file.write_text(TEST_GENERATE)
    test_file = str(test_file)

    gen_dir = tmp_path / "spam" / "egg"
    gen_dir.mkdir(parents=True)

    # If we don't generate, the test will fail
    assert_pytest_fails_with(['--mpl', test_file], 'Image file not found for comparison test')

    # If we do generate, the test should succeed and a new file will appear
    code = call_pytest([f'--mpl-generate-path={gen_dir}', test_file])
    assert code == 0
    assert os.path.exists(os.path.join(gen_dir, 'test_gen.png'))

    # If we do generate hash, the test will fail as no image is present
    hash_file = os.path.join(gen_dir, 'test_hashes.json')
    code = call_pytest([f'--mpl-generate-hash-library={hash_file}', test_file])
    assert code == 1
    assert os.path.exists(hash_file)

    with open(hash_file) as fp:
        hash_lib = json.load(fp)

    assert "test.test_gen" in hash_lib


@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_local, tolerance=20)
def test_tolerance():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 2])
    return fig


def test_nofigure():
    pass


@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_local,
                               style='fivethirtyeight',
                               tolerance=DEFAULT_TOLERANCE)
def test_base_style():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 3])
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_local,
                               remove_text=True)
def test_remove_text():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 3])
    return fig


@pytest.mark.parametrize('s', [5, 50, 500])
@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_local,
                               remove_text=True)
def test_parametrized(s):
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.scatter([1, 3, 4, 3, 2], [1, 4, 3, 3, 1], s=s)
    return fig


class TestClassWithSetup:

    # Regression test for a bug that occurred when using setup_method

    def setup_method(self, method):
        self.x = [1, 2, 3]

    @pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_local,
                                   filename='test_succeeds.png',
                                   tolerance=DEFAULT_TOLERANCE)
    def test_succeeds(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(self.x)
        return fig


class TestClassWithTestCase(TestCase):

    # Regression test for a bug that occurred when using unittest.TestCase

    def setUp(self):
        self.x = [1, 2, 3]

    @pytest.mark.mpl_image_compare(baseline_dir=baseline_dir_local,
                                   filename='test_succeeds.png',
                                   tolerance=DEFAULT_TOLERANCE)
    def test_succeeds(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(self.x)
        return fig


# hashlib

@pytest.mark.skipif(not hash_library.exists(), reason="No hash library for this mpl version")
@pytest.mark.mpl_image_compare(hash_library=hash_library, deterministic=True)
def test_hash_succeeds():
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 3])
    return fig


TEST_FAILING_HASH = rf"""
import pytest
import matplotlib.pyplot as plt
@pytest.mark.mpl_image_compare(hash_library=r"{fail_hash_library}", deterministic=True)
def test_hash_fails():
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot([1,2,2])
    return fig
"""


def test_hash_fails(tmp_path):
    test_file = tmp_path / "test.py"
    test_file.write_text(TEST_FAILING_HASH, encoding="ascii")
    test_file = str(test_file)

    # If we use --mpl, it should detect that the figure is wrong
    output = assert_pytest_fails_with(['--mpl', test_file], "doesn't match hash FAIL in library")
    # We didn't specify a baseline dir so we shouldn't attempt to find one
    assert "Image file not found for comparison test" not in output, output

    # Check that the summary path is printed and that it exists.
    output = assert_pytest_fails_with(['--mpl', test_file, '--mpl-generate-summary=html'],
                                      "doesn't match hash FAIL in library")
    # We didn't specify a baseline dir so we shouldn't attempt to find one
    print_message = "A summary of test results can be found at:"
    assert print_message in output, output
    printed_path = Path(output.split(print_message)[1].strip())
    assert printed_path.exists()

    # If we don't use --mpl option, the test should succeed
    code = call_pytest([test_file])
    assert code == 0


TEST_FAILING_HYBRID = rf"""
import pytest
import matplotlib.pyplot as plt
@pytest.mark.mpl_image_compare(hash_library=r"{fail_hash_library}",
                               tolerance=2, deterministic=True)
def test_hash_fail_hybrid():
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot([1,2,3])
    return fig
"""


@pytest.mark.skipif(ftv != '261', reason="Incorrect freetype version for hash check")
def test_hash_fail_hybrid(tmp_path):

    test_file = tmp_path / "test.py"
    test_file.write_text(TEST_FAILING_HYBRID, encoding="ascii")
    test_file = str(test_file)

    # Assert that image comparison runs and fails
    output = assert_pytest_fails_with(['--mpl', test_file,
                                       rf'--mpl-baseline-path={hash_baseline_dir_abs / "fail"}'],
                                      "doesn't match hash FAIL in library")
    assert "Error: Image files did not match." in output, output

    # Assert reports missing baseline image
    output = assert_pytest_fails_with(['--mpl', test_file,
                                       '--mpl-baseline-path=/not/a/path'],
                                      "doesn't match hash FAIL in library")
    assert "Image file not found for comparison test" in output, output

    # Assert reports image comparison succeeds
    output = assert_pytest_fails_with(['--mpl', test_file,
                                       rf'--mpl-baseline-path={hash_baseline_dir_abs / "succeed"}'],
                                      "doesn't match hash FAIL in library")
    assert "The comparison to the baseline image succeeded." in output, output

    # If we don't use --mpl option, the test should succeed
    code = call_pytest([test_file])
    assert code == 0


TEST_FAILING_NEW_HASH = r"""
import pytest
import matplotlib.pyplot as plt
@pytest.mark.mpl_image_compare
def test_hash_fails():
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot([1,2,2])
    return fig
"""


@pytest.mark.skipif(ftv != '261', reason="Incorrect freetype version for hash check")
def test_hash_fail_new_hashes(tmp_path):
    # Check that the hash comparison fails even if a new hash file is requested
    test_file = tmp_path / "test.py"
    test_file.write_text(TEST_FAILING_NEW_HASH, encoding="ascii")
    test_file = str(test_file)

    # Assert that image comparison runs and fails
    assert_pytest_fails_with(['--mpl', test_file,
                              f'--mpl-hash-library={fail_hash_library}'],
                             "doesn't match hash FAIL in library")

    hash_file = tmp_path / "new_hashes.json"
    # Assert that image comparison runs and fails
    assert_pytest_fails_with(['--mpl', test_file,
                              f'--mpl-hash-library={fail_hash_library}',
                              f'--mpl-generate-hash-library={hash_file}'],
                             "doesn't match hash FAIL")


TEST_MISSING_HASH = """
import pytest
import matplotlib.pyplot as plt
@pytest.mark.mpl_image_compare
def test_hash_missing():
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot([1,2,2])
    return fig
"""


def test_hash_missing(tmp_path):
    test_file = tmp_path / "test.py"
    test_file.write_text(TEST_MISSING_HASH)
    test_file = str(test_file)

    # Assert fails if hash library missing
    assert_pytest_fails_with(['--mpl', test_file, '--mpl-hash-library=/not/a/path'],
                             "Can't find hash library at path")

    # Assert fails if hash not in library
    assert_pytest_fails_with(['--mpl', test_file, f'--mpl-hash-library={fail_hash_library}'],
                             "Hash for test 'test.test_hash_missing' not found in")

    # If we don't use --mpl option, the test should succeed
    code = call_pytest([test_file])
    assert code == 0


TEST_RESULTS_ALWAYS = """
import pytest
import matplotlib.pyplot as plt
def plot():
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot([1,2,2])
    return fig
@pytest.mark.mpl_image_compare(deterministic=True)
def test_modified(): return plot()
@pytest.mark.mpl_image_compare(deterministic=True)
def test_new(): return plot()
@pytest.mark.mpl_image_compare(deterministic=True)
def test_unmodified(): return plot()
"""


@pytest.mark.skipif(not hash_library.exists(), reason="No hash library for this mpl version")
def test_results_always(tmp_path):
    test_file = tmp_path / "test.py"
    test_file.write_text(TEST_RESULTS_ALWAYS)
    results_path = tmp_path / "results"
    results_path.mkdir()

    code = call_pytest(['--mpl', str(test_file), '--mpl-results-always',
                        rf'--mpl-hash-library={hash_library}',
                        rf'--mpl-baseline-path={baseline_dir_abs}',
                        '--mpl-generate-summary=html,json,basic-html',
                        rf'--mpl-results-path={results_path}'])
    assert code == 0  # hashes correct, so all should pass

    # assert files for interactive HTML exist
    assert (results_path / "fig_comparison.html").exists()
    assert (results_path / "styles.css").exists()
    assert (results_path / "extra.js").exists()

    html = (results_path / "fig_comparison_basic.html").read_text()
    with (results_path / "results.json").open("r") as f:
        json_results = json.load(f)

    # each test, and which images should exist
    for test, exists in [
        ('test_modified', ['baseline', 'result-failed-diff', 'result']),
        ('test_new', ['result']),
        ('test_unmodified', ['baseline', 'result']),
    ]:

        test_name = f'test.{test}'

        summary = f'<div class="test-name">{test_name.split(".")[-1]}</div>'
        assert summary in html

        assert test_name in json_results.keys()
        json_res = json_results[test_name]
        assert json_res['status'] == 'passed'

        for image_type in ['baseline', 'result-failed-diff', 'result']:
            image = f'{test_name}/{image_type}.png'
            image_exists = (results_path / image).exists()
            json_image_key = f"{image_type.split('-')[-1]}_image"
            if image_type in exists:  # assert image so pytest prints it on error
                assert image and image_exists
                assert image in html
                assert json_res[json_image_key] == image
            else:
                assert image and not image_exists
                assert image not in html
                assert json_res[json_image_key] is None


TEST_FAILING_CLASS = """
import pytest
import matplotlib.pyplot as plt
class TestClass(object):
    @pytest.mark.mpl_image_compare
    def test_fails(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot([1, 2, 3])
        return fig
"""

TEST_FAILING_CLASS_SETUP_METHOD = """
import pytest
import matplotlib.pyplot as plt
class TestClassWithSetup:
    def setup_method(self, method):
        self.x = [1, 2, 3]
    @pytest.mark.mpl_image_compare
    def test_fails(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(self.x)
        return fig
"""

TEST_FAILING_UNITTEST_TESTCASE = """
from unittest import TestCase
import pytest
import matplotlib.pyplot as plt
class TestClassWithTestCase(TestCase):
    def setUp(self):
        self.x = [1, 2, 3]
    @pytest.mark.mpl_image_compare
    def test_fails(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(self.x)
        return fig
"""


@pytest.mark.parametrize("code", [
    TEST_FAILING_CLASS,
    TEST_FAILING_CLASS_SETUP_METHOD,
    TEST_FAILING_UNITTEST_TESTCASE,
])
def test_class_fail(code, tmp_path):

    test_file = tmp_path / "test.py"
    test_file.write_text(code)
    test_file = str(test_file)

    # Assert fails if hash library missing
    assert_pytest_fails_with(['--mpl', test_file, '--mpl-hash-library=/not/a/path'],
                             "Can't find hash library at path")

    # If we don't use --mpl option, the test should succeed
    code = call_pytest([test_file])
    assert code == 0


@pytest.mark.parametrize("runpytest_args", [(), ("--mpl",)])
def test_user_fail(pytester, runpytest_args):
    pytester.makepyfile(
        """
        import pytest
        @pytest.mark.mpl_image_compare
        def test_fail():
            pytest.fail("Manually failed by user.")
    """
    )
    result = pytester.runpytest(*runpytest_args)
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines("FAILED*Manually failed by user.*")


@pytest.mark.parametrize("runpytest_args", [(), ("--mpl",)])
def test_user_skip(pytester, runpytest_args):
    pytester.makepyfile(
        """
        import pytest
        @pytest.mark.mpl_image_compare
        def test_skip():
            pytest.skip("Manually skipped by user.")
    """
    )
    result = pytester.runpytest(*runpytest_args)
    result.assert_outcomes(skipped=1)


@pytest.mark.parametrize("runpytest_args", [(), ("--mpl",)])
def test_user_importorskip(pytester, runpytest_args):
    pytester.makepyfile(
        """
        import pytest
        @pytest.mark.mpl_image_compare
        def test_importorskip():
            pytest.importorskip("nonexistantmodule")
    """
    )
    result = pytester.runpytest(*runpytest_args)
    result.assert_outcomes(skipped=1)


@pytest.mark.parametrize("runpytest_args", [(), ("--mpl",)])
def test_user_xfail(pytester, runpytest_args):
    pytester.makepyfile(
        """
        import pytest
        @pytest.mark.mpl_image_compare
        def test_xfail():
            pytest.xfail()
    """
    )
    result = pytester.runpytest(*runpytest_args)
    result.assert_outcomes(xfailed=1)


@pytest.mark.parametrize("runpytest_args", [(), ("--mpl",)])
def test_user_exit_success(pytester, runpytest_args):
    pytester.makepyfile(
        """
        import pytest
        @pytest.mark.mpl_image_compare
        def test_exit_success():
            pytest.exit("Manually exited by user.", returncode=0)
    """
    )
    result = pytester.runpytest(*runpytest_args)
    result.assert_outcomes()
    assert result.ret == 0
    result.stdout.fnmatch_lines("*Exit*Manually exited by user.*")


@pytest.mark.parametrize("runpytest_args", [(), ("--mpl",)])
def test_user_exit_failure(pytester, runpytest_args):
    pytester.makepyfile(
        """
        import pytest
        @pytest.mark.mpl_image_compare
        def test_exit_fail():
            pytest.exit("Manually exited by user.", returncode=1)
    """
    )
    result = pytester.runpytest(*runpytest_args)
    result.assert_outcomes()
    assert result.ret == 1
    result.stdout.fnmatch_lines("*Exit*Manually exited by user.*")


@pytest.mark.parametrize("runpytest_args", [(), ("--mpl",)])
def test_user_function_raises(pytester, runpytest_args):
    pytester.makepyfile(
        """
        import pytest
        @pytest.mark.mpl_image_compare
        def test_raises():
            raise ValueError("User code raised an exception.")
    """
    )
    result = pytester.runpytest(*runpytest_args)
    result.assert_outcomes(failed=1)
    result.stdout.fnmatch_lines("FAILED*ValueError*User code*")


@pytest.mark.parametrize('use_hash_library', (False, True))
@pytest.mark.parametrize('passes', (False, True))
@pytest.mark.parametrize("file_format", ['eps', 'pdf', 'png', 'svg'])
def test_formats(pytester, use_hash_library, passes, file_format):
    """
    Note that we don't test all possible formats as some do not compress well
    and would bloat the baseline directory.
    """
    skip_if_format_unsupported(file_format, using_hashes=use_hash_library)
    if use_hash_library and not hash_library.exists():
        pytest.skip("No hash library for this mpl version")

    pytester.makepyfile(
        f"""
        import os
        import pytest
        import matplotlib.pyplot as plt
        @pytest.mark.mpl_image_compare(baseline_dir=r"{baseline_dir_abs}",
                                       {f'hash_library=r"{hash_library}",' if use_hash_library else ''}
                                       tolerance={DEFAULT_TOLERANCE},
                                       deterministic=True,
                                       savefig_kwargs={{'format': '{file_format}'}})
        def test_format_{file_format}():
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            ax.plot([{1 if passes else 3}, 2, 3])
            return fig
        """
    )
    result = pytester.runpytest('--mpl', '-rs')
    if passes:
        result.assert_outcomes(passed=1)
    else:
        result.assert_outcomes(failed=1)



import sys
from pathlib import Path

import matplotlib
import pytest
from matplotlib.testing.compare import converter
from packaging.version import Version

MPL_VERSION = Version(matplotlib.__version__)


def pytester_path(pytester):
    if hasattr(pytester, "path"):
        return pytester.path
    return Path(pytester.tmpdir)  # pytest v5


def skip_if_format_unsupported(file_format, using_hashes=False):
    if file_format == 'svg' and MPL_VERSION < Version('3.3'):
        pytest.skip('SVG comparison is only supported in Matplotlib 3.3 and above')

    if using_hashes:

        if file_format == 'pdf' and MPL_VERSION < Version('2.1'):
            pytest.skip('PDF hashes are only deterministic in Matplotlib 2.1 and above')
        elif file_format == 'eps' and MPL_VERSION < Version('2.1'):
            pytest.skip('EPS hashes are only deterministic in Matplotlib 2.1 and above')

    if using_hashes and not sys.platform.startswith('linux'):
        pytest.skip('Hashes for vector graphics are only provided in the hash library for Linux')

    if file_format != 'png' and file_format not in converter:
        if file_format == 'svg':
            pytest.skip('Comparing SVG files requires inkscape to be installed')
        else:
            pytest.skip('Comparing EPS and PDF files requires ghostscript to be installed')



import shutil
from pathlib import Path

import pytest
from helpers import pytester_path

FULL_TEST_NAME = "test_config.TestClass.test_mpl"
SHORT_TEST_NAME = "test_mpl"


@pytest.mark.parametrize(
    "ini, cli, expected_baseline_name, success_expected",
    [
        (None, None, SHORT_TEST_NAME, True),
        (False, None, SHORT_TEST_NAME, True),
        (True, None, FULL_TEST_NAME, True),
        (False, True, FULL_TEST_NAME, True),
        (None, True, FULL_TEST_NAME, True),
        (True, True, "bad_name", False),
    ],
)
def test_config(pytester, ini, cli, expected_baseline_name, success_expected):
    path = pytester_path(pytester)
    shutil.copyfile(  # Test will only pass if baseline is at expected path
        Path(__file__).parent / "baseline" / "2.0.x" / "test_base_style.png",
        path / f"{expected_baseline_name}.png",
    )
    ini = f"mpl-use-full-test-name = {ini}" if ini is not None else ""
    pytester.makeini(
        f"""
        [pytest]
        mpl-default-style = fivethirtyeight
        mpl-baseline-path = {path}
        {ini}
        """
    )
    pytester.makepyfile(
        """
        import matplotlib.pyplot as plt
        import pytest
        class TestClass:
            @pytest.mark.mpl_image_compare
            def test_mpl(self):
                fig, ax = plt.subplots()
                ax.plot([1, 2, 3])
                return fig
        """
    )
    cli = "--mpl-use-full-test-name" if cli else ""
    result = pytester.runpytest("--mpl", cli)
    if success_expected:
        result.assert_outcomes(passed=1)
    else:
        result.assert_outcomes(failed=1)



try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version

try:
    __version__ = version("pytest_mpl")
except PackageNotFoundError:
    __version__ = "unknown"



from pathlib import Path

import pytest
from PIL import Image, ImageDraw

TEST_NAME = "test_base_style"


@pytest.fixture(scope="module")
def baseline_image(tmp_path_factory):
    path = Path(__file__).parent / "baseline" / "2.0.x" / f"{TEST_NAME}.png"
    with Image.open(path) as image:
        draw = ImageDraw.Draw(image)
        draw.rectangle(((0, 0), (100, 100)), fill="red")
        output = tmp_path_factory.mktemp("data") / f"{TEST_NAME}.png"
        image.save(output)
    return output


@pytest.mark.parametrize(
    "ini, cli, kwarg, success_expected",
    [
        (40, None, None, True),
        (30, 40, None, True),
        (30, 30, 40, True),
        (30, 40, 30, False),
        (40, 30, 30, False),
    ],
)
def test_config(pytester, baseline_image, ini, cli, kwarg, success_expected):
    ini = f"mpl-default-tolerance = {ini}" if ini else ""
    pytester.makeini(
        f"""
        [pytest]
        mpl-default-style = fivethirtyeight
        mpl-baseline-path = {baseline_image.parent}
        {ini}
        """
    )
    kwarg = f"tolerance={kwarg}" if kwarg else ""
    pytester.makepyfile(
        f"""
        import matplotlib.pyplot as plt
        import pytest
        @pytest.mark.mpl_image_compare({kwarg})
        def {TEST_NAME}():
            fig, ax = plt.subplots()
            ax.plot([1, 2, 3])
            return fig
        """
    )
    cli = f"--mpl-default-tolerance={cli}" if cli else ""
    result = pytester.runpytest("--mpl", cli)
    if success_expected:
        result.assert_outcomes(passed=1)
    else:
        result.assert_outcomes(failed=1)
